"""
Module for running vmadm command on SmartOS
"""

import logging
import os
import shlex
from collections import OrderedDict

import salt.utils.args
import salt.utils.files
import salt.utils.json
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils

log = logging.getLogger(__name__)

# Function aliases
__func_alias__ = {"list_vms": "list"}

# Define the module's virtual name
__virtualname__ = "vmadm"


def __virtual__():
    """
    Provides vmadm on SmartOS
    """
    if (
        salt.utils.platform.is_smartos_globalzone()
        and salt.utils.path.which("vmadm")
        and salt.utils.path.which("zfs")
    ):
        return __virtualname__
    return (
        False,
        f"{__virtualname__} module can only be loaded on SmartOS compute nodes",
    )


def _exit_status(retcode):
    """
    Translate exit status of vmadm
    """
    if retcode == 0:
        return "Successful completion."
    if retcode == 1:
        return "An error occurred."
    if retcode == 2:
        return "Usage error."


def _create_update_from_file(mode="create", uuid=None, path=None):
    """
    Create vm from file
    """
    ret = {}
    if not os.path.isfile(path) or path is None:
        ret["Error"] = f"File ({path}) does not exists!"
        return ret
    # vmadm validate create|update [-f <filename>]
    cmd = "vmadm validate {mode} {brand} -f {path}".format(
        mode=mode, brand=get(uuid)["brand"] if uuid is not None else "", path=path
    )
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        if "stderr" in res:
            if res["stderr"][0] == "{":
                ret["Error"] = salt.utils.json.loads(res["stderr"])
            else:
                ret["Error"] = res["stderr"]
        return ret
    # vmadm create|update [-f <filename>]
    cmd = "vmadm {mode} {uuid} -f {path}".format(
        mode=mode, uuid=uuid if uuid is not None else "", path=path
    )
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        if "stderr" in res:
            if res["stderr"][0] == "{":
                ret["Error"] = salt.utils.json.loads(res["stderr"])
            else:
                ret["Error"] = res["stderr"]
        return ret
    else:
        if res["stderr"].startswith("Successfully created VM"):
            return res["stderr"][24:]
    return True


def _create_update_from_cfg(mode="create", uuid=None, vmcfg=None):
    """
    Create vm from configuration
    """
    ret = {}

    # write json file
    vmadm_json_file = __salt__["temp.file"](prefix="vmadm-")
    with salt.utils.files.fopen(vmadm_json_file, "w") as vmadm_json:
        salt.utils.json.dump(vmcfg, vmadm_json)

    # vmadm validate create|update [-f <filename>]
    cmd = "vmadm validate {mode} {brand} -f {vmadm_json_file}".format(
        mode=mode,
        brand=get(uuid)["brand"] if uuid is not None else "",
        vmadm_json_file=vmadm_json_file,
    )
    res = __salt__["cmd.run_all"](cmd, python_shell=True)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        if "stderr" in res:
            if res["stderr"][0] == "{":
                ret["Error"] = salt.utils.json.loads(res["stderr"])
            else:
                ret["Error"] = res["stderr"]
        return ret
    # vmadm create|update [-f <filename>]
    cmd = "vmadm {mode} {uuid} -f {vmadm_json_file}".format(
        mode=mode,
        uuid=uuid if uuid is not None else "",
        vmadm_json_file=vmadm_json_file,
    )
    res = __salt__["cmd.run_all"](cmd, python_shell=True)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        if "stderr" in res:
            if res["stderr"][0] == "{":
                ret["Error"] = salt.utils.json.loads(res["stderr"])
            else:
                ret["Error"] = res["stderr"]
        return ret
    else:
        # cleanup json file (only when successful to help troubleshooting)
        salt.utils.files.safe_rm(vmadm_json_file)

        # return uuid
        if res["stderr"].startswith("Successfully created VM"):
            return res["stderr"][24:]

    return True


def start(vm, options=None, key="uuid"):
    """
    Start a vm

    vm : string
        vm to be started
    options : string
        optional additional options
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.start 186da9ab-7392-4f55-91a5-b8f1fe770543
        salt '*' vmadm.start 186da9ab-7392-4f55-91a5-b8f1fe770543 'order=c,once=d cdrom=/path/to/image.iso,ide'
        salt '*' vmadm.start vm=nacl key=alias
        salt '*' vmadm.start vm=nina.example.org key=hostname
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup(f"{key}={vm}", one=True)
    if "Error" in vm:
        return vm
    # vmadm start <uuid> [option=value ...]
    cmd = "vmadm start {uuid} {options}".format(
        uuid=vm, options=options if options else ""
    )
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def stop(vm, force=False, key="uuid"):
    """
    Stop a vm

    vm : string
        vm to be stopped
    force : boolean
        force stop of vm if true
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.stop 186da9ab-7392-4f55-91a5-b8f1fe770543
        salt '*' vmadm.stop 186da9ab-7392-4f55-91a5-b8f1fe770543 True
        salt '*' vmadm.stop vm=nacl key=alias
        salt '*' vmadm.stop vm=nina.example.org key=hostname
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup(f"{key}={vm}", one=True)
    if "Error" in vm:
        return vm
    # vmadm stop <uuid> [-F]
    cmd = "vmadm stop {force} {uuid}".format(force="-F" if force else "", uuid=vm)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        return ret
    return True


def reboot(vm, force=False, key="uuid"):
    """
    Reboot a vm

    vm : string
        vm to be rebooted
    force : boolean
        force reboot of vm if true
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.reboot 186da9ab-7392-4f55-91a5-b8f1fe770543
        salt '*' vmadm.reboot 186da9ab-7392-4f55-91a5-b8f1fe770543 True
        salt '*' vmadm.reboot vm=nacl key=alias
        salt '*' vmadm.reboot vm=nina.example.org key=hostname
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup(f"{key}={vm}", one=True)
    if "Error" in vm:
        return vm
    # vmadm reboot <uuid> [-F]
    cmd = "vmadm reboot {force} {uuid}".format(force="-F" if force else "", uuid=vm)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def list_vms(search=None, sort=None, order="uuid,type,ram,state,alias", keyed=True):
    """
    Return a list of VMs

    search : string
        vmadm filter property
    sort : string
        vmadm sort (-s) property
    order : string
        vmadm order (-o) property -- Default: uuid,type,ram,state,alias
    keyed : boolean
        specified if the output should be an array (False) or dict (True)
            For a dict the key is the first item from the order parameter.
            Note: If key is not unique last vm wins.

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.list
        salt '*' vmadm.list order=alias,ram,cpu_cap sort=-ram,-cpu_cap
        salt '*' vmadm.list search='type=KVM'
    """
    ret = {}
    # vmadm list [-p] [-H] [-o field,...] [-s field,...] [field=value ...]
    cmd = "vmadm list -p -H {order} {sort} {search}".format(
        order=f"-o {order}" if order else "",
        sort=f"-s {sort}" if sort else "",
        search=search if search else "",
    )
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    result = OrderedDict() if keyed else []
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret

    fields = order.split(",")

    for vm in res["stdout"].splitlines():
        vm_data = OrderedDict()
        vm = vm.split(":")
        if keyed:
            for field in fields:
                if fields.index(field) == 0:
                    continue
                vm_data[field.strip()] = vm[fields.index(field)].strip()
            result[vm[0]] = vm_data
        else:
            if len(vm) > 1:
                for field in fields:
                    vm_data[field.strip()] = vm[fields.index(field)].strip()
            else:
                vm_data = vm[0]
            result.append(vm_data)
    return result


def lookup(search=None, order=None, one=False):
    """
    Return a list of VMs using lookup

    search : string
        vmadm filter property
    order : string
        vmadm order (-o) property -- Default: uuid,type,ram,state,alias
    one : boolean
        return only one result (vmadm's -1)

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.lookup search='state=running'
        salt '*' vmadm.lookup search='state=running' order=uuid,alias,hostname
        salt '*' vmadm.lookup search='alias=nacl' one=True
    """
    ret = {}
    # vmadm lookup [-j|-1] [-o field,...] [field=value ...]
    cmd = "vmadm lookup {one} {order} {search}".format(
        one="-1" if one else "-j",
        order=f"-o {order}" if order else "",
        search=search if search else "",
    )
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    result = []
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret

    if one:
        result = res["stdout"]
    else:
        for vm in salt.utils.json.loads(res["stdout"]):
            result.append(vm)

    return result


def sysrq(vm, action="nmi", key="uuid"):
    """
    Send non-maskable interrupt to vm or capture a screenshot

    vm : string
        vm to be targeted
    action : string
        nmi or screenshot -- Default: nmi
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.sysrq 186da9ab-7392-4f55-91a5-b8f1fe770543 nmi
        salt '*' vmadm.sysrq 186da9ab-7392-4f55-91a5-b8f1fe770543 screenshot
        salt '*' vmadm.sysrq nacl nmi key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    if action not in ["nmi", "screenshot"]:
        ret["Error"] = "Action must be either nmi or screenshot"
        return ret
    vm = lookup(f"{key}={vm}", one=True)
    if "Error" in vm:
        return vm
    # vmadm sysrq <uuid> <nmi|screenshot>
    cmd = f"vmadm sysrq {vm} {action}"
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def delete(vm, key="uuid"):
    """
    Delete a vm

    vm : string
        vm to be deleted
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.delete 186da9ab-7392-4f55-91a5-b8f1fe770543
        salt '*' vmadm.delete nacl key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup(f"{key}={vm}", one=True)
    if "Error" in vm:
        return vm
    # vmadm delete <uuid>
    cmd = f"vmadm delete {vm}"
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def get(vm, key="uuid"):
    """
    Output the JSON object describing a VM

    vm : string
        vm to be targeted
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.get 186da9ab-7392-4f55-91a5-b8f1fe770543
        salt '*' vmadm.get nacl key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup(f"{key}={vm}", one=True)
    if "Error" in vm:
        return vm
    # vmadm get <uuid>
    cmd = f"vmadm get {vm}"
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return salt.utils.json.loads(res["stdout"])


def info(vm, info_type="all", key="uuid"):
    """
    Lookup info on running kvm

    vm : string
        vm to be targeted
    info_type : string [all|block|blockstats|chardev|cpus|kvm|pci|spice|version|vnc]
        info type to return
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.info 186da9ab-7392-4f55-91a5-b8f1fe770543
        salt '*' vmadm.info 186da9ab-7392-4f55-91a5-b8f1fe770543 vnc
        salt '*' vmadm.info nacl key=alias
        salt '*' vmadm.info nacl vnc key=alias
    """
    ret = {}
    if info_type not in [
        "all",
        "block",
        "blockstats",
        "chardev",
        "cpus",
        "kvm",
        "pci",
        "spice",
        "version",
        "vnc",
    ]:
        ret["Error"] = "Requested info_type is not available"
        return ret
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup(f"{key}={vm}", one=True)
    if "Error" in vm:
        return vm
    # vmadm info <uuid> [type,...]
    cmd = f"vmadm info {vm} {info_type}"
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return salt.utils.json.loads(res["stdout"])


def create_snapshot(vm, name, key="uuid"):
    """
    Create snapshot of a vm

    vm : string
        vm to be targeted
    name : string
        snapshot name
            The snapname must be 64 characters or less
            and must only contain alphanumeric characters and
            characters in the set [-_.:%] to comply with ZFS restrictions.
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.create_snapshot 186da9ab-7392-4f55-91a5-b8f1fe770543 baseline
        salt '*' vmadm.create_snapshot nacl baseline key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup(f"{key}={vm}", one=True)
    if "Error" in vm:
        return vm
    vmobj = get(vm)
    if "datasets" in vmobj:
        ret["Error"] = "VM cannot have datasets"
        return ret
    if vmobj["brand"] in ["kvm"]:
        ret["Error"] = "VM must be of type OS"
        return ret
    if vmobj["zone_state"] not in ["running"]:  # work around a vmadm bug
        ret["Error"] = "VM must be running to take a snapshot"
        return ret
    # vmadm create-snapshot <uuid> <snapname>
    cmd = f"vmadm create-snapshot {vm} {name}"
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def delete_snapshot(vm, name, key="uuid"):
    """
    Delete snapshot of a vm

    vm : string
        vm to be targeted
    name : string
        snapshot name
            The snapname must be 64 characters or less
            and must only contain alphanumeric characters and
            characters in the set [-_.:%] to comply with ZFS restrictions.
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.delete_snapshot 186da9ab-7392-4f55-91a5-b8f1fe770543 baseline
        salt '*' vmadm.delete_snapshot nacl baseline key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup(f"{key}={vm}", one=True)
    if "Error" in vm:
        return vm
    vmobj = get(vm)
    if "datasets" in vmobj:
        ret["Error"] = "VM cannot have datasets"
        return ret
    if vmobj["brand"] in ["kvm"]:
        ret["Error"] = "VM must be of type OS"
        return ret
    # vmadm delete-snapshot <uuid> <snapname>
    cmd = f"vmadm delete-snapshot {vm} {name}"
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def rollback_snapshot(vm, name, key="uuid"):
    """
    Rollback snapshot of a vm

    vm : string
        vm to be targeted
    name : string
        snapshot name
            The snapname must be 64 characters or less
            and must only contain alphanumeric characters and
            characters in the set [-_.:%] to comply with ZFS restrictions.
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.rollback_snapshot 186da9ab-7392-4f55-91a5-b8f1fe770543 baseline
        salt '*' vmadm.rollback_snapshot nacl baseline key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup(f"{key}={vm}", one=True)
    if "Error" in vm:
        return vm
    vmobj = get(vm)
    if "datasets" in vmobj:
        ret["Error"] = "VM cannot have datasets"
        return ret
    if vmobj["brand"] in ["kvm"]:
        ret["Error"] = "VM must be of type OS"
        return ret
    # vmadm rollback-snapshot <uuid> <snapname>
    cmd = f"vmadm rollback-snapshot {vm} {name}"
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def reprovision(vm, image, key="uuid"):
    """
    Reprovision a vm

    vm : string
        vm to be reprovisioned
    image : string
        uuid of new image
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.reprovision 186da9ab-7392-4f55-91a5-b8f1fe770543 c02a2044-c1bd-11e4-bd8c-dfc1db8b0182
        salt '*' vmadm.reprovision nacl c02a2044-c1bd-11e4-bd8c-dfc1db8b0182 key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup(f"{key}={vm}", one=True)
    if "Error" in vm:
        return vm
    if image not in __salt__["imgadm.list"]():
        ret["Error"] = f"Image ({image}) is not present on this host"
        return ret
    # vmadm reprovision <uuid> [-f <filename>]
    cmd = "echo {image} | vmadm reprovision {uuid}".format(
        uuid=salt.utils.stringutils.to_unicode(vm),
        image=shlex.quote(salt.utils.json.dumps({"image_uuid": image})),
    )
    res = __salt__["cmd.run_all"](cmd, python_shell=True)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def create(from_file=None, **kwargs):
    """
    Create a new vm

    from_file : string
        json file to create the vm from -- if present, all other options will be ignored
    kwargs : string|int|...
        options to set for the vm

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.create from_file=/tmp/new_vm.json
        salt '*' vmadm.create image_uuid='...' alias='...' nics='[{ "nic_tag": "admin", "ip": "198.51.100.123", ...}, {...}]' [...]
    """
    ret = {}
    # prepare vmcfg
    vmcfg = {}
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    for k, v in kwargs.items():
        vmcfg[k] = v

    if from_file:
        return _create_update_from_file("create", path=from_file)
    else:
        return _create_update_from_cfg("create", vmcfg=vmcfg)


def update(vm, from_file=None, key="uuid", **kwargs):
    """
    Update a new vm

    vm : string
        vm to be updated
    from_file : string
        json file to update the vm with -- if present, all other options will be ignored
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter
    kwargs : string|int|...
        options to update for the vm

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.update vm=186da9ab-7392-4f55-91a5-b8f1fe770543 from_file=/tmp/new_vm.json
        salt '*' vmadm.update vm=nacl key=alias from_file=/tmp/new_vm.json
        salt '*' vmadm.update vm=186da9ab-7392-4f55-91a5-b8f1fe770543 max_physical_memory=1024
    """
    ret = {}
    # prepare vmcfg
    vmcfg = {}
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    for k, v in kwargs.items():
        vmcfg[k] = v

    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    uuid = lookup(f"{key}={vm}", one=True)
    if "Error" in uuid:
        return uuid

    if from_file:
        return _create_update_from_file("update", uuid, path=from_file)
    else:
        return _create_update_from_cfg("update", uuid, vmcfg=vmcfg)


def send(vm, target, key="uuid"):
    """
    Send a vm to a directory

    vm : string
        vm to be sent
    target : string
        target directory
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.send 186da9ab-7392-4f55-91a5-b8f1fe770543 /opt/backups
        salt '*' vmadm.send vm=nacl target=/opt/backups key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    if not os.path.isdir(target):
        ret["Error"] = "Target must be a directory or host"
        return ret
    vm = lookup(f"{key}={vm}", one=True)
    if "Error" in vm:
        return vm
    # vmadm send <uuid> [target]
    cmd = "vmadm send {uuid} > {target}".format(
        uuid=vm, target=os.path.join(target, f"{vm}.vmdata")
    )
    res = __salt__["cmd.run_all"](cmd, python_shell=True)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    vmobj = get(vm)
    if "datasets" not in vmobj:
        return True
    log.warning("one or more datasets detected, this is not supported!")
    log.warning("trying to zfs send datasets...")
    for dataset in vmobj["datasets"]:
        name = dataset.split("/")
        name = name[-1]
        cmd = "zfs send {dataset} > {target}".format(
            dataset=dataset,
            target=os.path.join(target, f"{vm}-{name}.zfsds"),
        )
        res = __salt__["cmd.run_all"](cmd, python_shell=True)
        retcode = res["retcode"]
        if retcode != 0:
            ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
            return ret
    return True


def receive(uuid, source):
    """
    Receive a vm from a directory

    uuid : string
        uuid of vm to be received
    source : string
        source directory

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.receive 186da9ab-7392-4f55-91a5-b8f1fe770543 /opt/backups
    """
    ret = {}
    if not os.path.isdir(source):
        ret["Error"] = "Source must be a directory or host"
        return ret
    if not os.path.exists(os.path.join(source, f"{uuid}.vmdata")):
        ret["Error"] = f"Unknow vm with uuid in {source}"
        return ret
    # vmadm receive
    cmd = "vmadm receive < {source}".format(
        source=os.path.join(source, f"{uuid}.vmdata")
    )
    res = __salt__["cmd.run_all"](cmd, python_shell=True)
    retcode = res["retcode"]
    if retcode != 0 and not res["stderr"].endswith("datasets"):
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    vmobj = get(uuid)
    if "datasets" not in vmobj:
        return True
    log.warning("one or more datasets detected, this is not supported!")
    log.warning("trying to restore datasets, mountpoints will need to be set again...")
    for dataset in vmobj["datasets"]:
        name = dataset.split("/")
        name = name[-1]
        cmd = "zfs receive {dataset} < {source}".format(
            dataset=dataset,
            source=os.path.join(source, f"{uuid}-{name}.zfsds"),
        )
        res = __salt__["cmd.run_all"](cmd, python_shell=True)
        retcode = res["retcode"]
        if retcode != 0:
            ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
            return ret
    cmd = f"vmadm install {uuid}"
    res = __salt__["cmd.run_all"](cmd, python_shell=True)
    retcode = res["retcode"]
    if retcode != 0 and not res["stderr"].endswith("datasets"):
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>

    Unit tests for the salt.modules.cmdmod module
"""

import builtins
import getpass
import logging
import os
import re
import sys
import tempfile

import pytest

import salt.grains.extra
import salt.modules.cmdmod as cmdmod
import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
from salt._logging import LOG_LEVELS
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, Mock, MockTimedProc, mock_open, patch
from tests.support.runtests import RUNTIME_VARS

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
]

DEFAULT_SHELL = "foo/bar"
MOCK_SHELL_FILE = "# List of acceptable shells\n\n/bin/bash\n"


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {cmdmod: {"__opts__": minion_opts}}


@pytest.fixture(scope="module")
def mock_loglevels():
    return {
        "info": "foo",
        "all": "bar",
        "critical": "bar",
        "trace": "bar",
        "garbage": "bar",
        "error": "bar",
        "debug": "bar",
        "warning": "bar",
        "quiet": "bar",
    }


def test_render_cmd_no_template():
    """
    Tests return when template=None
    """
    assert cmdmod._render_cmd("foo", "bar", None) == ("foo", "bar")


def test_render_cmd_saltenv_from_config():
    mock_template = MagicMock()
    with patch.dict(cmdmod.__opts__, {"saltenv": "base"}):
        with patch.dict(
            "salt.utils.templates.TEMPLATE_REGISTRY", {"test": mock_template}
        ):
            cmdmod._render_cmd("test", "test", "test")
            assert mock_template.call_count == 2
            assert mock_template.call_args[1]["saltenv"] == "base"
            cmdmod._render_cmd("test", "test", "test", saltenv="fake")
            assert mock_template.call_count == 4
            assert mock_template.call_args[1]["saltenv"] == "fake"


def test_render_cmd_unavailable_engine():
    """
    Tests CommandExecutionError raised when template isn't in the
    template registry
    """
    with pytest.raises(CommandExecutionError):
        cmdmod._render_cmd("boo", "bar", "baz")


def test_check_loglevel_bad_level(mock_loglevels):
    """
    Tests return of providing an invalid loglevel option
    """
    with patch.dict(LOG_LEVELS, mock_loglevels):
        assert cmdmod._check_loglevel(level="bad_loglevel") == "foo"


def test_check_loglevel_bad_level_not_str(mock_loglevels):
    """
    Tests the return of providing an invalid loglevel option that is not a string
    """
    with patch.dict(LOG_LEVELS, mock_loglevels):
        assert cmdmod._check_loglevel(level=1000) == "foo"


def test_check_loglevel_quiet(mock_loglevels):
    """
    Tests the return of providing a loglevel of 'quiet'
    """
    with patch.dict(LOG_LEVELS, mock_loglevels):
        assert cmdmod._check_loglevel(level="quiet") is None


def test_parse_env_not_env():
    """
    Tests the return of an env that is not an env
    """
    assert cmdmod._parse_env(None) == {}


def test_parse_env_list():
    """
    Tests the return of an env that is a list
    """
    ret = {"foo": None, "bar": None}
    assert ret == cmdmod._parse_env(["foo", "bar"])


def test_parse_env_dict():
    """
    Test the return of an env that is not a dict
    """
    assert cmdmod._parse_env("test") == {}


def test_run_shell_is_not_file():
    """
    Tests error raised when shell is not available after _is_valid_shell error msg
    and os.path.isfile returns False
    """
    with patch("salt.modules.cmdmod._is_valid_shell", MagicMock(return_value=True)):
        with patch("salt.utils.platform.is_windows", MagicMock(return_value=False)):
            with patch("os.path.isfile", MagicMock(return_value=False)):
                with pytest.raises(CommandExecutionError):
                    cmdmod._run("foo", "bar")


def test_run_shell_file_no_access():
    """
    Tests error raised when shell is not available after _is_valid_shell error msg,
    os.path.isfile returns True, but os.access returns False
    """
    with patch("salt.modules.cmdmod._is_valid_shell", MagicMock(return_value=True)):
        with patch("salt.utils.platform.is_windows", MagicMock(return_value=False)):
            with patch("os.path.isfile", MagicMock(return_value=True)):
                with patch("os.access", MagicMock(return_value=False)):
                    with pytest.raises(CommandExecutionError):
                        cmdmod._run("foo", "bar")


def test_run_runas_with_windows():
    """
    Tests error raised when runas is passed on windows
    """
    with patch("salt.modules.cmdmod._is_valid_shell", MagicMock(return_value=True)):
        with patch("salt.utils.platform.is_windows", MagicMock(return_value=True)):
            with patch(
                "salt.utils.win_chcp.get_codepage_id", MagicMock(return_value=65001)
            ):
                with patch.dict(cmdmod.__grains__, {"os": "fake_os"}):
                    with pytest.raises(CommandExecutionError):
                        cmdmod._run("foo", "bar", runas="baz")


def test_run_with_tuple():
    """
    Tests return when cmd is a tuple
    """
    mock_true = MagicMock(return_value=True)
    with patch("salt.modules.cmdmod._is_valid_shell", mock_true):
        with patch("salt.utils.platform.is_windows", MagicMock(return_value=False)):
            with patch("os.path.isfile", mock_true):
                with patch("os.access", mock_true):
                    cmdmod._run(("echo", "foo"), python_shell=True, cwd="/")


def test_run_user_not_available():
    """
    Tests return when runas user is not available
    """
    mock_true = MagicMock(return_value=True)
    with patch("salt.modules.cmdmod._is_valid_shell", mock_true):
        with patch("os.path.isfile", mock_true):
            with patch("os.access", mock_true):
                with pytest.raises(CommandExecutionError):
                    cmdmod._run("foo", "bar", runas="baz")


def test_run_zero_umask():
    """
    Tests error raised when umask is set to zero
    """
    with patch("salt.modules.cmdmod._is_valid_shell", MagicMock(return_value=True)):
        with patch("salt.utils.platform.is_windows", MagicMock(return_value=False)):
            with patch("os.path.isfile", MagicMock(return_value=True)):
                with patch("os.access", MagicMock(return_value=True)):
                    with pytest.raises(CommandExecutionError):
                        cmdmod._run("foo", "bar", umask=0)


def test_run_invalid_umask():
    """
    Tests error raised when an invalid umask is given
    """
    with patch("salt.modules.cmdmod._is_valid_shell", MagicMock(return_value=True)):
        with patch("salt.utils.platform.is_windows", MagicMock(return_value=False)):
            with patch("os.path.isfile", MagicMock(return_value=True)):
                with patch("os.access", MagicMock(return_value=True)):
                    pytest.raises(
                        CommandExecutionError,
                        cmdmod._run,
                        "foo",
                        "bar",
                        umask="baz",
                    )


def test_run_invalid_cwd_not_abs_path():
    """
    Tests error raised when cwd is not an absolute path
    """
    with patch("salt.modules.cmdmod._is_valid_shell", MagicMock(return_value=True)):
        with patch("salt.utils.platform.is_windows", MagicMock(return_value=False)):
            with patch("os.path.isfile", MagicMock(return_value=True)):
                with patch("os.access", MagicMock(return_value=True)):
                    with pytest.raises(CommandExecutionError):
                        cmdmod._run("foo", "bar")


def test_run_invalid_cwd_not_dir():
    """
    Tests error raised when cwd is not a dir
    """
    with patch("salt.modules.cmdmod._is_valid_shell", MagicMock(return_value=True)):
        with patch("salt.utils.platform.is_windows", MagicMock(return_value=False)):
            with patch("os.path.isfile", MagicMock(return_value=True)):
                with patch("os.access", MagicMock(return_value=True)):
                    with patch("os.path.isabs", MagicMock(return_value=True)):
                        with pytest.raises(CommandExecutionError):
                            cmdmod._run("foo", "bar")


def test_run_no_vt_os_error():
    """
    Tests error raised when not useing vt and OSError is provided
    """
    expected_error = "expect error"
    with patch("salt.modules.cmdmod._is_valid_shell", MagicMock(return_value=True)):
        with patch("salt.utils.platform.is_windows", MagicMock(return_value=False)):
            with patch("os.path.isfile", MagicMock(return_value=True)):
                with patch("os.access", MagicMock(return_value=True)):
                    with patch(
                        "salt.utils.timed_subprocess.TimedProc",
                        MagicMock(side_effect=OSError(expected_error)),
                    ):
                        with pytest.raises(CommandExecutionError) as error:
                            cmdmod.run("foo", cwd="/")
                        assert error.value.args[0].endswith(expected_error)


def test_run_no_vt_io_error():
    """
    Tests error raised when not useing vt and IOError is provided
    """
    expected_error = "expect error"
    with patch("salt.modules.cmdmod._is_valid_shell", MagicMock(return_value=True)):
        with patch("salt.utils.platform.is_windows", MagicMock(return_value=False)):
            with patch("os.path.isfile", MagicMock(return_value=True)):
                with patch("os.access", MagicMock(return_value=True)):
                    with patch(
                        "salt.utils.timed_subprocess.TimedProc",
                        MagicMock(side_effect=IOError(expected_error)),
                    ):
                        with pytest.raises(CommandExecutionError) as error:
                            cmdmod.run("foo", cwd="/")
                        assert error.value.args[0].endswith(expected_error)


@pytest.mark.skip(reason="Test breaks unittests runs")
@pytest.mark.skip_on_windows
def test_run():
    """
    Tests end result when a command is not found
    """
    with patch("salt.modules.cmdmod._is_valid_shell", MagicMock(return_value=True)):
        with patch("salt.utils.platform.is_windows", MagicMock(return_value=False)):
            with patch("os.path.isfile", MagicMock(return_value=True)):
                with patch("os.access", MagicMock(return_value=True)):
                    ret = cmdmod._run("foo", cwd=os.getcwd(), use_vt=True).get("stderr")
                    assert "foo" in ret


@pytest.mark.skip_unless_on_windows
def test_powershell():
    """
    Tests cmd.powershell with a string value output
    """
    mock_run = {"pid": 1234, "retcode": 0, "stderr": "", "stdout": '"foo"'}
    with patch("salt.modules.cmdmod._run", return_value=mock_run):
        ret = cmdmod.powershell("Set-ExecutionPolicy RemoteSigned")
        assert ret == "foo"


@pytest.mark.skip_unless_on_windows
def test_powershell_empty():
    """
    Tests cmd.powershell when the output is an empty string
    """
    mock_run = {"pid": 1234, "retcode": 0, "stderr": "", "stdout": ""}
    with patch("salt.modules.cmdmod._run", return_value=mock_run):
        ret = cmdmod.powershell("Set-ExecutionPolicy RemoteSigned")
        assert ret == ""


def test_is_valid_shell_windows():
    """
    Tests return if running on windows
    """
    with patch("salt.utils.platform.is_windows", MagicMock(return_value=True)):
        assert cmdmod._is_valid_shell("foo")


@pytest.mark.skip_on_windows
def test_is_valid_shell_none():
    """
    Tests return of when os.path.exists(/etc/shells) isn't available
    """
    with patch("os.path.exists", MagicMock(return_value=False)):
        assert cmdmod._is_valid_shell("foo") is None


def test_is_valid_shell_available():
    """
    Tests return when provided shell is available
    """
    with patch("os.path.exists", MagicMock(return_value=True)):
        with patch("salt.utils.files.fopen", mock_open(read_data=MOCK_SHELL_FILE)):
            assert cmdmod._is_valid_shell("/bin/bash")


@pytest.mark.skip_on_windows
def test_is_valid_shell_unavailable():
    """
    Tests return when provided shell is not available
    """
    with patch("os.path.exists", MagicMock(return_value=True)):
        with patch("salt.utils.files.fopen", mock_open(read_data=MOCK_SHELL_FILE)):
            assert not cmdmod._is_valid_shell("foo")


@pytest.mark.skip_on_windows
def test_os_environment_remains_intact():
    """
    Make sure the OS environment is not tainted after running a command
    that specifies runas.
    """
    with patch("pwd.getpwnam") as getpwnam_mock:
        with patch("subprocess.Popen") as popen_mock:
            environment = os.environ.copy()

            popen_mock.return_value = Mock(
                communicate=lambda *args, **kwags: [b"", None],
                pid=lambda: 1,
                retcode=0,
            )

            with patch.dict(
                cmdmod.__grains__, {"os": "Darwin", "os_family": "Solaris"}
            ):
                if sys.platform.startswith(("freebsd", "openbsd")):
                    shell = "/bin/sh"
                else:
                    shell = "/bin/bash"

                cmdmod._run(
                    "ls", cwd=tempfile.gettempdir(), runas="foobar", shell=shell
                )

                environment2 = os.environ.copy()

                assert environment == environment2

                if not salt.utils.platform.is_darwin():
                    getpwnam_mock.assert_called_with("foobar")


@pytest.mark.skip_on_windows
def test_os_environment_do_not_pass_notify_socket():
    """
    Make sure NOTIFY_SOCKET environment variable is not passed
    to the command if not explicitly set with env parameter.
    """
    with patch("pwd.getpwnam") as getpwnam_mock:
        new_env = os.environ.copy()
        new_env.update({"NOTIFY_SOCKET": "/run/systemd/notify"})
        with patch("subprocess.Popen") as popen_mock, patch(
            "os.environ.copy", return_value=new_env
        ):
            popen_mock.return_value = Mock(
                communicate=lambda *args, **kwags: [b"", None],
                pid=lambda: 1,
                retcode=0,
            )

            with patch.dict(cmdmod.__grains__, {"os": "SUSE", "os_family": "Suse"}):
                if sys.platform.startswith(("freebsd", "openbsd")):
                    shell = "/bin/sh"
                else:
                    shell = "/bin/bash"

                cmdmod._run("ls", cwd=tempfile.gettempdir(), shell=shell)

                assert "NOTIFY_SOCKET" not in popen_mock.call_args_list[0][1]["env"]

                cmdmod._run(
                    "ls",
                    cwd=tempfile.gettempdir(),
                    shell=shell,
                    env={"NOTIFY_SOCKET": "/run/systemd/notify.new"},
                )

                assert (
                    popen_mock.call_args_list[1][1]["env"]["NOTIFY_SOCKET"]
                    == "/run/systemd/notify.new"
                )


@pytest.mark.skip_unless_on_darwin
def test_shell_properly_handled_on_macOS():
    """
    cmd.run should invoke a new bash login only
    when bash is the default shell for the selected user
    """

    class _CommandHandler:
        """
        Class for capturing cmd
        """

        def __init__(self):
            self.cmd = None

        def clear(self):
            self.cmd = None

    cmd_handler = _CommandHandler()

    def mock_proc(__cmd__, **kwargs):
        cmd_handler.cmd = " ".join(__cmd__)
        return MagicMock(return_value=MockTimedProc(stdout=None, stderr=None))

    with patch("pwd.getpwnam") as getpwnam_mock:
        with patch("salt.utils.timed_subprocess.TimedProc", mock_proc):

            # User default shell is '/usr/local/bin/bash'
            user_default_shell = "/usr/local/bin/bash"
            with patch(
                "pwd.getpwall",
                Mock(
                    return_value=[Mock(pw_shell=user_default_shell, pw_name="foobar")]
                ),
            ):
                cmd_handler.clear()
                cmdmod._run(
                    "ls", cwd=tempfile.gettempdir(), runas="foobar", use_vt=False
                )

                assert re.search(
                    f"{user_default_shell} -l -c", cmd_handler.cmd
                ), "cmd invokes right bash session on macOS"

            # User default shell is '/bin/zsh'
            user_default_shell = "/bin/zsh"
            with patch(
                "pwd.getpwall",
                Mock(
                    return_value=[Mock(pw_shell=user_default_shell, pw_name="foobar")]
                ),
            ):

                cmd_handler.clear()
                cmdmod._run(
                    "ls", cwd=tempfile.gettempdir(), runas="foobar", use_vt=False
                )

                assert not re.search(
                    "bash -l -c", cmd_handler.cmd
                ), "cmd does not invoke user shell on macOS"


@pytest.mark.skip_on_windows
def test_run_all_quiet_does_not_depend_on_salt_dunder():
    """
    `cmdmod._run_all_quiet` should not depend on availability
    of __salt__ dictionary (issue #61816).

    This test checks for __salt__ specifically and will still
    pass if other dunders, especially __grains__, are referenced.
    This is the case on UNIX systems other than MacOS when
    `sudo` could not be found.
    """

    proc = MagicMock(return_value=MockTimedProc(stdout=b"success", stderr=None))
    runas = getpass.getuser()

    with patch.dict(cmdmod.__grains__, {"os": "Darwin", "os_family": "Solaris"}):
        with patch("salt.utils.timed_subprocess.TimedProc", proc):
            salt_dunder_mock = MagicMock(spec_set=dict)
            salt_dunder_mock.__getitem__.side_effect = NameError(
                "__salt__ might not be defined"
            )

            with patch.object(cmdmod, "__salt__", salt_dunder_mock):
                ret = cmdmod._run_all_quiet("foo")
                assert ret["stdout"] == "success"
                assert salt_dunder_mock.__getitem__.call_count == 0
                ret = cmdmod._run_all_quiet("foo", runas=runas)
                assert ret["stdout"] == "success"
                assert salt_dunder_mock.__getitem__.call_count == 0


def test_run_cwd_doesnt_exist_issue_7154():
    """
    cmd.run should fail and raise
    salt.exceptions.CommandExecutionError if the cwd dir does not
    exist
    """
    cmd = "echo OHAI"
    cwd = "/path/to/nowhere"
    with pytest.raises(CommandExecutionError):
        cmdmod.run_all(cmd, cwd=cwd)


@pytest.mark.skip_on_darwin
@pytest.mark.skip_on_windows
def test_run_cwd_in_combination_with_runas():
    """
    cmd.run executes command in the cwd directory
    when the runas parameter is specified
    """
    cmd = "pwd"
    cwd = "/tmp"
    runas = getpass.getuser()

    with patch.dict(cmdmod.__grains__, {"os": "Darwin", "os_family": "Solaris"}):
        stdout = cmdmod._run(cmd, cwd=cwd, runas=runas).get("stdout")
    assert stdout == cwd


def test_run_all_binary_replace():
    """
    Test for failed decoding of binary data, for instance when doing
    something silly like using dd to read from /dev/urandom and write to
    /dev/stdout.
    """
    # Since we're using unicode_literals, read the random bytes from a file
    rand_bytes_file = os.path.join(RUNTIME_VARS.BASE_FILES, "random_bytes")
    with salt.utils.files.fopen(rand_bytes_file, "rb") as fp_:
        stdout_bytes = fp_.read()

    stdout_bytes = stdout_bytes.rstrip() + os.linesep.encode()

    # stdout with the non-decodable bits replaced with the unicode
    # replacement character U+FFFD.
    stdout_unicode = "\ufffd\x1b\ufffd\ufffd" + os.linesep
    stderr_bytes = (
        os.linesep.encode().join(
            [
                b"1+0 records in",
                b"1+0 records out",
                b"4 bytes copied, 9.1522e-05 s, 43.7 kB/s",
            ]
        )
        + os.linesep.encode()
    )
    stderr_unicode = stderr_bytes.decode()

    proc = MagicMock(
        return_value=MockTimedProc(stdout=stdout_bytes, stderr=stderr_bytes)
    )
    with patch("salt.utils.timed_subprocess.TimedProc", proc):
        ret = cmdmod.run_all(
            "dd if=/dev/urandom of=/dev/stdout bs=4 count=1", rstrip=False
        )

    assert ret["stdout"] == stdout_unicode
    assert ret["stderr"] == stderr_unicode


def test_run_all_none():
    """
    Tests cases when proc.stdout or proc.stderr are None. These should be
    caught and replaced with empty strings.
    """
    proc = MagicMock(return_value=MockTimedProc(stdout=None, stderr=None))
    with patch("salt.utils.timed_subprocess.TimedProc", proc):
        ret = cmdmod.run_all("some command", rstrip=False)

    assert ret["stdout"] == ""
    assert ret["stderr"] == ""


def test_run_all_unicode():
    """
    Ensure that unicode stdout and stderr are decoded properly
    """
    stdout_unicode = "Here is some unicode: спам"
    stderr_unicode = "Here is some unicode: яйца"
    stdout_bytes = stdout_unicode.encode("utf-8")
    stderr_bytes = stderr_unicode.encode("utf-8")

    proc = MagicMock(
        return_value=MockTimedProc(stdout=stdout_bytes, stderr=stderr_bytes)
    )

    with patch("salt.utils.timed_subprocess.TimedProc", proc), patch.object(
        builtins, "__salt_system_encoding__", "utf-8"
    ):
        ret = cmdmod.run_all("some command", rstrip=False)

    assert ret["stdout"] == stdout_unicode
    assert ret["stderr"] == stderr_unicode


def test_run_all_output_encoding():
    """
    Test that specifying the output encoding works as expected
    """
    stdout = "Æ"
    stdout_latin1_enc = stdout.encode("latin1")

    proc = MagicMock(return_value=MockTimedProc(stdout=stdout_latin1_enc))

    with patch("salt.utils.timed_subprocess.TimedProc", proc), patch.object(
        builtins, "__salt_system_encoding__", "utf-8"
    ):
        ret = cmdmod.run_all("some command", output_encoding="latin1")

    assert ret["stdout"] == stdout


def test_run_all_output_loglevel_quiet(caplog):
    """
    Test that specifying quiet for loglevel
    does not log the command.
    """
    stdout = b"test"
    proc = MagicMock(return_value=MockTimedProc(stdout=stdout))

    msg = "Executing command 'some command' in directory"
    with patch("salt.utils.timed_subprocess.TimedProc", proc):
        with caplog.at_level(logging.DEBUG, logger="salt.modules.cmdmod"):
            ret = cmdmod.run_all("some command", output_loglevel="quiet")
        assert msg not in caplog.text

    assert ret["stdout"] == salt.utils.stringutils.to_unicode(stdout)


def test_run_all_output_loglevel_debug(caplog):
    """
    Test that specifying debug for loglevel
    does log the command.
    """
    stdout = b"test"
    proc = MagicMock(return_value=MockTimedProc(stdout=stdout))

    # When we get back to having to specify a shell, we may need to change this
    # back.
    # if salt.utils.platform.is_windows():
    #     run_cmd = salt.utils.path.which("cmd")
    #     expected = f"Executing command '{run_cmd}' in directory"
    # else:
    #     expected = "Executing command 'some' in directory"
    expected = "Executing command 'some' in directory"

    with patch("salt.utils.timed_subprocess.TimedProc", proc):
        with caplog.at_level(logging.DEBUG, logger="salt.modules.cmdmod"):
            ret = cmdmod.run_all("some command", output_loglevel="debug")
        result = caplog.text
        assert expected.lower() in result.lower()

    assert ret["stdout"] == salt.utils.stringutils.to_unicode(stdout)


def test_run_chroot_mount():
    """
    Test cmdmod.run_chroot mount / umount balance
    """
    mock_mount = MagicMock()
    mock_umount = MagicMock()
    mock_run_all = MagicMock()
    with patch.dict(
        cmdmod.__salt__, {"mount.mount": mock_mount, "mount.umount": mock_umount}
    ):
        with patch("salt.modules.cmdmod.run_all", mock_run_all):
            cmdmod.run_chroot("/mnt", "cmd")
            assert mock_mount.call_count == 3
            assert mock_umount.call_count == 3


def test_run_chroot_mount_bind():
    """
    Test cmdmod.run_chroot mount / umount balance with bind mount
    """
    mock_mount = MagicMock()
    mock_umount = MagicMock()
    mock_run_all = MagicMock()
    with patch.dict(
        cmdmod.__salt__, {"mount.mount": mock_mount, "mount.umount": mock_umount}
    ):
        with patch("salt.modules.cmdmod.run_all", mock_run_all):
            cmdmod.run_chroot("/mnt", "cmd", binds=["/var"])
            assert mock_mount.call_count == 4
            assert mock_umount.call_count == 4


@pytest.mark.skip_on_windows
def test_run_chroot_runas():
    """
    Test run_chroot when a runas parameter is provided
    """
    with patch.dict(
        cmdmod.__salt__, {"mount.mount": MagicMock(), "mount.umount": MagicMock()}
    ):
        with patch("salt.modules.cmdmod.run_all") as run_all_mock:
            cmdmod.run_chroot("/mnt", "ls", runas="foobar", shell="/bin/sh")
    run_all_mock.assert_called_with(
        "chroot --userspec foobar: /mnt /bin/sh -c ls",
        bg=False,
        clean_env=False,
        cwd=None,
        env=None,
        ignore_retcode=False,
        log_callback=None,
        output_encoding=None,
        output_loglevel="quiet",
        pillar=None,
        pillarenv=None,
        python_shell=True,
        reset_system_locale=True,
        rstrip=True,
        saltenv=None,
        shell="/bin/sh",
        stdin=None,
        success_retcodes=None,
        success_stdout=None,
        success_stderr=None,
        template=None,
        timeout=None,
        umask=None,
        use_vt=False,
    )


def test_cve_2021_25284(caplog):
    proc = MagicMock(
        return_value=MockTimedProc(stdout=b"foo", stderr=b"wtf", returncode=2)
    )
    with patch("salt.utils.timed_subprocess.TimedProc", proc):
        with caplog.at_level(logging.DEBUG, logger="salt.modules.cmdmod"):
            cmdmod.run("testcmd -p ImAPassword", output_loglevel="error")
        assert "ImAPassword" not in caplog.text


def test__log_cmd_str():
    "_log_cmd function handles strings"
    assert cmdmod._log_cmd("foo bar") == "foo"


def test__log_cmd_list():
    "_log_cmd function handles lists"
    assert cmdmod._log_cmd(["foo", "bar"]) == "foo"


def test_log_cmd_tuple():
    "_log_cmd function handles tuples"
    assert cmdmod._log_cmd(("foo", "bar")) == "foo"


def test_log_cmd_non_str_tuple_list():
    "_log_cmd function casts objects to strings"

    class cmd:
        def __init__(self, cmd):
            self.cmd = cmd

        def __str__(self):
            return self.cmd

    assert cmdmod._log_cmd(cmd("foo bar")) == "foo"


@pytest.mark.skip_on_windows
def test_cmd_script_saltenv_from_config():
    mock_cp_get_template = MagicMock()
    mock_cp_cache_file = MagicMock()
    mock_run = MagicMock()
    with patch.dict(cmdmod.__opts__, {"saltenv": "base"}):
        with patch.dict(
            cmdmod.__salt__,
            {
                "cp.cache_file": mock_cp_cache_file,
                "cp.get_template": mock_cp_get_template,
                "file.user_to_uid": MagicMock(),
                "file.remove": MagicMock(),
            },
        ):
            with patch("salt.modules.cmdmod._run") as mock_run:
                with patch("shutil.copyfile", MagicMock()):
                    with patch("os.chmod", MagicMock()):
                        with patch("os.chown", MagicMock()):
                            cmdmod.script("test")
                            assert mock_cp_cache_file.call_count == 1
                            mock_cp_cache_file.assert_called_with("test", "base")
                            assert mock_run.call_count == 1
                            assert mock_run.call_args[1]["saltenv"] == "base"
                            cmdmod.script("test", template="jinja")
                            assert mock_cp_get_template.call_count == 1
                            assert mock_cp_get_template.call_args[0][3] == "base"
                            assert mock_run.call_count == 2
                            assert mock_run.call_args[1]["saltenv"] == "base"


@pytest.mark.skip_unless_on_windows
def test_cmd_script_saltenv_from_config_windows():
    mock_cp_get_template = MagicMock()
    mock_cp_cache_file = MagicMock()
    with patch.dict(cmdmod.__opts__, {"saltenv": "base"}):
        with patch.dict(
            cmdmod.__salt__,
            {
                "cp.cache_file": mock_cp_cache_file,
                "cp.get_template": mock_cp_get_template,
                "file.remove": MagicMock(),
            },
        ):
            with patch("salt.modules.cmdmod._run") as mock_run:
                with patch("shutil.copyfile", MagicMock()):
                    cmdmod.script("test")
                    assert mock_cp_cache_file.call_count == 1
                    mock_cp_cache_file.assert_called_with("test", "base")
                    assert mock_run.call_count == 1
                    assert mock_run.call_args[1]["saltenv"] == "base"
                    cmdmod.script("test", template="jinja")
                    assert mock_cp_get_template.call_count == 1
                    assert mock_cp_get_template.call_args[0][3] == "base"
                    assert mock_run.call_count == 2
                    assert mock_run.call_args[1]["saltenv"] == "base"


@pytest.mark.parametrize("bundled", [True, False])
@pytest.mark.parametrize(
    "test_os,test_family",
    [
        ("FreeBSD", "FreeBSD"),
        ("linux", "Solaris"),
        ("linux", "AIX"),
        ("linux", "linux"),
    ],
)
@pytest.mark.skip_on_darwin
@pytest.mark.skip_on_windows
def test_runas_env_all_os(test_os, test_family, bundled):
    """
    cmd.run executes command and the environment is returned
    when the runas parameter is specified
    on all different OS types and os_family
    """
    with patch("pwd.getpwnam") as getpwnam_mock:
        with patch("subprocess.Popen") as popen_mock:
            popen_mock.return_value = Mock(
                communicate=lambda *args, **kwags: [b"", None],
                pid=lambda: 1,
                retcode=0,
            )
            file_name = "/tmp/doesnotexist"

            with patch.dict(
                cmdmod.__grains__, {"os": test_os, "os_family": test_family}
            ):
                with patch("salt.utils.pkg.check_bundled", return_value=bundled):
                    with patch("shutil.chown"):
                        with patch("os.remove"):
                            with patch.object(
                                tempfile, "NamedTemporaryFile"
                            ) as mock_fp:
                                mock_fp.return_value.__enter__.return_value.name = (
                                    file_name
                                )
                                if sys.platform.startswith(("freebsd", "openbsd")):
                                    shell = "/bin/sh"
                                else:
                                    shell = "/bin/bash"
                                _user = "foobar"
                                cmdmod._run(
                                    "ls",
                                    cwd=tempfile.gettempdir(),
                                    runas=_user,
                                    shell=shell,
                                )
                                if not bundled:
                                    if test_family in ("Solaris", "AIX"):
                                        env_cmd = ["su", "-", _user, "-c"]
                                    elif test_os == "FreeBSD":
                                        env_cmd = ["su", "-", _user, "-c"]
                                    else:
                                        env_cmd = [
                                            "su",
                                            "-s",
                                            shell,
                                            "-",
                                            _user,
                                            "-c",
                                        ]
                                    if test_os == "FreeBSD":
                                        env_cmd.extend([f"{shell} -c {sys.executable}"])
                                    else:
                                        env_cmd.extend([sys.executable])
                                    assert popen_mock.call_args_list[0][0][0] == env_cmd
                                else:
                                    if test_family in ("Solaris", "AIX"):
                                        env_cmd = ["su", "-", _user, "-c"]
                                    elif test_os == "FreeBSD":
                                        env_cmd = ["su", "-", _user, "-c"]
                                    else:
                                        env_cmd = [
                                            "su",
                                            "-s",
                                            shell,
                                            "-",
                                            _user,
                                            "-c",
                                        ]
                                    if test_os == "FreeBSD":
                                        env_cmd.extend(
                                            [
                                                "{} -c {} python {}".format(
                                                    shell, sys.executable, file_name
                                                )
                                            ]
                                        )
                                    else:
                                        env_cmd.extend(
                                            [
                                                "{} python {}".format(
                                                    sys.executable, file_name
                                                )
                                            ]
                                        )
                                    assert popen_mock.call_args_list[0][0][0] == env_cmd


@pytest.mark.skip_on_darwin
@pytest.mark.skip_on_windows
@pytest.mark.parametrize("bundled", [True, False])
def test_runas_env_sudo_group(bundled):
    """
    cmd.run executes command and the environment is returned
    when the runas parameter is specified
    when group is passed and use_sudo=True
    """
    with patch("pwd.getpwnam") as getpwnam_mock:
        with patch("subprocess.Popen") as popen_mock:
            popen_mock.return_value = Mock(
                communicate=lambda *args, **kwags: [b"", None],
                pid=lambda: 1,
                retcode=0,
            )
            file_name = "/tmp/doesnotexist"

            with patch.dict(cmdmod.__grains__, {"os": "linux", "os_family": "linux"}):
                with patch("grp.getgrnam"):
                    with patch("salt.utils.pkg.check_bundled", return_value=bundled):
                        with patch("shutil.chown"):
                            with patch("os.remove"):
                                with patch.object(
                                    tempfile, "NamedTemporaryFile"
                                ) as mock_fp:
                                    mock_fp.return_value.__enter__.return_value.name = (
                                        file_name
                                    )
                                    if sys.platform.startswith(("freebsd", "openbsd")):
                                        shell = "/bin/sh"
                                    else:
                                        shell = "/bin/bash"
                                    _user = "foobar"
                                    _group = "foobar"
                                    same_shell = False
                                    if salt.grains.extra.shell()["shell"] == shell:
                                        same_shell = True

                                    cmdmod._run(
                                        "ls",
                                        cwd=tempfile.gettempdir(),
                                        runas=_user,
                                        shell=shell,
                                        group=_group,
                                    )
                                    if not bundled:
                                        exp_ret = [
                                            "sudo",
                                            "-u",
                                            _user,
                                            "-g",
                                            _group,
                                            "-s",
                                            "--",
                                            shell,
                                            "-c",
                                            sys.executable,
                                        ]
                                        if same_shell:
                                            exp_ret = [
                                                "sudo",
                                                "-u",
                                                _user,
                                                "-g",
                                                _group,
                                                "-i",
                                                "--",
                                                sys.executable,
                                            ]
                                        assert (
                                            popen_mock.call_args_list[0][0][0]
                                            == exp_ret
                                        )
                                    else:
                                        exp_ret = [
                                            "sudo",
                                            "-u",
                                            _user,
                                            "-g",
                                            _group,
                                            "-s",
                                            "--",
                                            shell,
                                            "-c",
                                            "{} python {}".format(
                                                sys.executable, file_name
                                            ),
                                        ]
                                        if same_shell:
                                            exp_ret = [
                                                "sudo",
                                                "-u",
                                                _user,
                                                "-g",
                                                _group,
                                                "-i",
                                                "--",
                                                "{} python {}".format(
                                                    sys.executable, file_name
                                                ),
                                            ]
                                        assert (
                                            popen_mock.call_args_list[0][0][0]
                                            == exp_ret
                                        )


@pytest.mark.skip_unless_on_windows
def test__run_no_powershell():
    with pytest.raises(CommandExecutionError):
        cmdmod._run(shell="unk_bin", cmd="Some-Command", encoded_cmd=False)


@pytest.mark.parametrize(
    "cmd, parsed",
    [
        ("Write-Host foo", "Write-Host foo"),
        ("& Write-Host foo", "& Write-Host foo"),
        ("$PSVersionTable", "$PSVersionTable"),
        ("try {this} catch {that}", "try {this} catch {that}"),
        ("[bool]@{value = 0}", "[bool]@{value = 0}"),
        (
            "(Get-Date(Get-Date).ToUniversalTime() -UFormat %s)",
            "(Get-Date(Get-Date).ToUniversalTime() -UFormat %s)",
        ),
        (
            "if (1 -eq 1) { exit 0 } else { exit 1 }",
            "if (1 -eq 1) { exit 0 } else { exit 1 }",
        ),
        (
            "do { $count++; $a++; } while ($x[$a] -ne 0)",
            "do { $count++; $a++; } while ($x[$a] -ne 0)",
        ),
        (
            "while ($val -ne 3) { $val++; Write-Host $val }",
            "while ($val -ne 3) { $val++; Write-Host $val }",
        ),
        (
            "trap { 'Error found.' }",
            "trap { 'Error found.' }",
        ),
        (
            "for ($i=1; $i -le 10; $i++) { Write-Host $i }",
            "for ($i=1; $i -le 10; $i++) { Write-Host $i }",
        ),
        (
            "foreach ($file in Get-ChildItem) { Write-Host $file }",
            "foreach ($file in Get-ChildItem) { Write-Host $file }",
        ),
        (
            'data { if ($null) { "To get help for this cmdlet, type Get-Help New-Dictionary." } }',
            'data { if ($null) { "To get help for this cmdlet, type Get-Help New-Dictionary." } }',
        ),
    ],
)
@pytest.mark.skip_unless_on_windows
def test_prep_powershell_cmd(cmd, parsed):
    """
    Tests _prep_powershell_cmd returns correct cmd
    """
    ret = cmdmod._prep_powershell_cmd(
        win_shell="powershell.exe", cmd=cmd, encoded_cmd=False
    )
    expected = [
        "powershell.exe",
        "-NonInteractive",
        "-NoProfile",
        "-ExecutionPolicy",
        "Bypass",
        "-Command",
        parsed,
    ]
    assert ret == expected


@pytest.mark.skip_unless_on_windows
def test_prep_powershell_cmd_encoded():
    """
    Tests _prep_powershell_cmd returns correct cmd when encoded_cmd=True
    """
    # This is the encoded command for 'Write-Host "Encoded HOLO"'
    e_cmd = "VwByAGkAdABlAC0ASABvAHMAdAAgACIARQBuAGMAbwBkAGUAZAAgAEgATwBMAE8AIgA="
    ret = cmdmod._prep_powershell_cmd(
        win_shell="powershell.exe", cmd=e_cmd, encoded_cmd=True
    )
    expected = [
        "powershell.exe",
        "-NonInteractive",
        "-NoProfile",
        "-ExecutionPolicy",
        "Bypass",
        "-EncodedCommand",
        e_cmd,
    ]
    assert ret == expected


@pytest.mark.skip_unless_on_windows
def test_prep_powershell_cmd_script():
    """
    Tests _prep_powershell_cmd returns correct cmd when called from cmd.script
    """
    stack = [["", "", ""], ["", "", "script"], ["", "", ""], ["", "", ""]]
    with patch("traceback.extract_stack", return_value=stack), patch(
        "salt.utils.path.which", return_value="powershell.exe"
    ):
        script = r"C:\some\script.ps1"
        ret = cmdmod._prep_powershell_cmd(
            win_shell="powershell.exe", cmd=[script], encoded_cmd=False
        )
        expected = [
            "powershell.exe",
            "-NonInteractive",
            "-NoProfile",
            "-ExecutionPolicy",
            "Bypass",
            "-File",
            script,
        ]
        assert ret == expected


@pytest.mark.parametrize(
    "text, expected",
    [
        ("", '""'),  # Should quote an empty string
        ("Foo", '"Foo"'),  # Should quote a string
        ('["foo", "bar"]', '["foo", "bar"]'),  # Should leave unchanged
        ('{"foo": "bar"}', '{"foo": "bar"}'),  # Should leave unchanged
    ],
)
@pytest.mark.skip_unless_on_windows
def test_prep_powershell_json(text, expected):
    """
    Make sure the output is valid json
    """
    result = cmdmod._prep_powershell_json(text)
    assert result == expected



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.djangomod
"""

import pytest

import salt.modules.djangomod as djangomod
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    with patch("salt.utils.path.which", lambda exe: exe):
        yield {djangomod: {}}


def test_command():
    """
    Test if it runs arbitrary django management command
    """
    mock = MagicMock(return_value=True)
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        assert djangomod.command("DJANGO_SETTINGS_MODULE", "validate")


def test_syncdb():
    """
    Test if it runs the Django-Admin syncdb command
    """
    mock = MagicMock(return_value=True)
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        assert djangomod.syncdb("DJANGO_SETTINGS_MODULE")


def test_migrate():
    """
    Test if it runs the Django-Admin migrate command
    """
    mock = MagicMock(return_value=True)
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        assert djangomod.migrate("DJANGO_SETTINGS_MODULE")


def test_createsuperuser():
    """
    Test if it create a super user for the database.
    """
    mock = MagicMock(return_value=True)
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        assert djangomod.createsuperuser(
            "DJANGO_SETTINGS_MODULE", "SALT", "salt@slatstack.com"
        )


def test_loaddata():
    """
    Test if it loads fixture data
    """
    mock = MagicMock(return_value=True)
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        assert djangomod.loaddata("DJANGO_SETTINGS_MODULE", "mydata")


def test_collectstatic():
    """
    Test if it collect static files from each of your applications
    into a single location
    """
    mock = MagicMock(return_value=True)
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        assert djangomod.collectstatic("DJANGO_SETTINGS_MODULE")


def test_django_admin_cli_command():
    mock = MagicMock()
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        djangomod.command("settings.py", "runserver")
        mock.assert_called_once_with(
            "django-admin.py runserver --settings=settings.py",
            python_shell=False,
            env=None,
            runas=None,
        )


def test_django_admin_cli_command_with_args():
    mock = MagicMock()
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        djangomod.command(
            "settings.py",
            "runserver",
            None,
            None,
            None,
            None,
            "noinput",
            "somethingelse",
        )
        mock.assert_called_once_with(
            "django-admin.py runserver --settings=settings.py "
            "--noinput --somethingelse",
            python_shell=False,
            env=None,
            runas=None,
        )


def test_django_admin_cli_command_with_kwargs():
    mock = MagicMock()
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        djangomod.command(
            "settings.py", "runserver", None, None, None, database="something"
        )
        mock.assert_called_once_with(
            "django-admin.py runserver --settings=settings.py --database=something",
            python_shell=False,
            env=None,
            runas=None,
        )


def test_django_admin_cli_command_with_kwargs_ignore_dunder():
    mock = MagicMock()
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        djangomod.command(
            "settings.py", "runserver", None, None, None, __ignore="something"
        )
        mock.assert_called_once_with(
            "django-admin.py runserver --settings=settings.py",
            python_shell=False,
            env=None,
            runas=None,
        )


def test_django_admin_cli_syncdb():
    mock = MagicMock()
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        djangomod.syncdb("settings.py")
        mock.assert_called_once_with(
            "django-admin.py syncdb --settings=settings.py --noinput",
            python_shell=False,
            env=None,
            runas=None,
        )


def test_django_admin_cli_syncdb_migrate():
    mock = MagicMock()
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        djangomod.syncdb("settings.py", migrate=True)
        mock.assert_called_once_with(
            "django-admin.py syncdb --settings=settings.py --migrate --noinput",
            python_shell=False,
            env=None,
            runas=None,
        )


def test_django_admin_cli_migrate():
    mock = MagicMock()
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        djangomod.migrate("settings.py")
        mock.assert_called_once_with(
            "django-admin.py migrate --settings=settings.py --noinput",
            python_shell=False,
            env=None,
            runas=None,
        )


def test_django_admin_cli_createsuperuser():
    mock = MagicMock()
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        djangomod.createsuperuser("settings.py", "testuser", "user@example.com")
        assert mock.call_count == 1
        mock.assert_called_with(
            "django-admin.py createsuperuser --settings=settings.py --noinput "
            "--email=user@example.com --username=testuser",
            env=None,
            python_shell=False,
            runas=None,
        )


def no_test_loaddata():
    mock = MagicMock()
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        djangomod.loaddata("settings.py", "app1,app2")
        mock.assert_called_once_with(
            "django-admin.py loaddata --settings=settings.py app1 app2",
        )


def test_django_admin_cli_collectstatic():
    mock = MagicMock()
    with patch.dict(djangomod.__salt__, {"cmd.run": mock}):
        djangomod.collectstatic(
            "settings.py", None, True, "something", True, True, True, True
        )
        mock.assert_called_once_with(
            "django-admin.py collectstatic --settings=settings.py "
            "--noinput --no-post-process --dry-run --clear --link "
            "--no-default-ignore --ignore=something",
            python_shell=False,
            env=None,
            runas=None,
        )



"""
Functional tests for authentication version downgrade attacks.

These tests demonstrate the vulnerability where a malicious minion can
downgrade its authentication protocol version to bypass security features.
"""

import ctypes
import multiprocessing
import pathlib

import pytest

import salt.channel.client
import salt.channel.server
import salt.config
import salt.crypt
import salt.daemons.masterapi
import salt.ext.tornado.gen
import salt.ext.tornado.ioloop
import salt.master
import salt.payload
import salt.utils.event
import salt.utils.files
import salt.utils.platform
import salt.utils.process
import salt.utils.stringutils
from salt.master import SMaster

pytestmark = [
    pytest.mark.skip_on_spawning_platform(
        reason="These tests are currently broken on spawning platforms. Need to be rewritten.",
    ),
    pytest.mark.timeout_unless_on_windows(120),
]


@pytest.fixture
def channel_minion_id():
    return "test-minion-downgrade"


@pytest.fixture
def auth_pki_dir(tmp_path):
    """Setup PKI directory structure for functional auth tests."""
    if salt.utils.platform.is_darwin():
        # To avoid 'OSError: AF_UNIX path too long'
        _root_dir = pathlib.Path("/tmp").resolve() / tmp_path.name
        root_dir = _root_dir
    else:
        root_dir = tmp_path

    master_pki = root_dir / "master_pki"
    minion_pki = root_dir / "minion_pki"

    master_pki.mkdir(parents=True)
    minion_pki.mkdir(parents=True)
    (master_pki / "minions").mkdir()
    (master_pki / "minions_pre").mkdir()
    (master_pki / "minions_rejected").mkdir()
    (master_pki / "minions_denied").mkdir()

    # Generate keys
    salt.crypt.gen_keys(str(master_pki), "master", 4096)
    salt.crypt.gen_keys(str(minion_pki), "minion", 4096)

    yield root_dir

    if salt.utils.platform.is_darwin():
        import shutil

        shutil.rmtree(str(root_dir), ignore_errors=True)


def _create_functional_master_opts(
    auth_pki_dir, channel_minion_id, minimum_auth_version=0
):
    """Helper to create master configuration with specified minimum_auth_version."""
    opts = {
        "master_uri": "tcp://127.0.0.1:44506",
        "interface": "127.0.0.1",
        "ret_port": 44506,
        "ipv6": False,
        "sock_dir": str(auth_pki_dir / "sock"),
        "cachedir": str(auth_pki_dir / "cache"),
        "pki_dir": str(auth_pki_dir / "master_pki"),
        "id": "master",
        "__role": "master",
        "transport": "zeromq",
        "keysize": 4096,
        "max_minions": 0,
        "auto_accept": True,
        "open_mode": False,
        "key_pass": None,
        "publish_port": 44505,
        "auth_mode": 1,
        "auth_events": True,
        "publish_session": 86400,
        "request_server_ttl": 300,  # 5 minutes
        "worker_threads": 1,
        "cluster_id": None,
        "master_sign_pubkey": False,
        "sign_pub_messages": False,
        "minimum_auth_version": minimum_auth_version,
    }
    (auth_pki_dir / "sock").mkdir(exist_ok=True)
    (auth_pki_dir / "cache").mkdir(exist_ok=True)
    (auth_pki_dir / "cache" / "sessions").mkdir(exist_ok=True)

    # Accept the minion key
    minion_pub = auth_pki_dir / "minion_pki" / "minion.pub"
    master_minions = auth_pki_dir / "master_pki" / "minions" / channel_minion_id
    if minion_pub.exists():
        with salt.utils.files.fopen(str(minion_pub)) as src:
            with salt.utils.files.fopen(str(master_minions), "w") as dst:
                dst.write(src.read())

    return opts


@pytest.fixture
def functional_master_opts(auth_pki_dir, channel_minion_id):
    """Master configuration for functional tests (default: minimum_auth_version=3)."""
    return _create_functional_master_opts(
        auth_pki_dir, channel_minion_id, minimum_auth_version=3
    )


@pytest.fixture
def functional_minion_opts(auth_pki_dir, functional_master_opts, channel_minion_id):
    """Minion configuration for functional tests."""
    return {
        "master_uri": "tcp://127.0.0.1:44506",
        "interface": "127.0.0.1",
        "ret_port": 44506,
        "ipv6": False,
        "sock_dir": str(functional_master_opts["sock_dir"]),
        "pki_dir": str(auth_pki_dir / "minion_pki"),
        "id": channel_minion_id,
        "__role": "minion",
        "transport": "zeromq",
        "keysize": 4096,
        "encryption_algorithm": salt.crypt.OAEP_SHA1,
        "signing_algorithm": salt.crypt.PKCS1v15_SHA1,
        "master_port": 44506,
        "master_ip": "127.0.0.1",
    }


@pytest.mark.parametrize(
    "minimum_auth_version,attack_version,should_pass",
    [
        pytest.param(
            0,
            0,
            False,
            marks=pytest.mark.xfail(
                reason="Vulnerable: minimum_auth_version=0 allows all versions",
                strict=True,
            ),
        ),
        pytest.param(
            0,
            1,
            False,
            marks=pytest.mark.xfail(
                reason="Vulnerable: minimum_auth_version=0 allows all versions",
                strict=True,
            ),
        ),
        pytest.param(
            0,
            2,
            False,
            marks=pytest.mark.xfail(
                reason="Vulnerable: minimum_auth_version=0 allows all versions",
                strict=True,
            ),
        ),
        pytest.param(1, 0, True, id="v1-rejects-v0"),
        pytest.param(
            1,
            1,
            False,
            marks=pytest.mark.xfail(
                reason="Vulnerable: minimum_auth_version=1 allows v1", strict=True
            ),
        ),
        pytest.param(
            1,
            2,
            False,
            marks=pytest.mark.xfail(
                reason="Vulnerable: minimum_auth_version=1 allows v2", strict=True
            ),
        ),
        pytest.param(2, 0, True, id="v2-rejects-v0"),
        pytest.param(2, 1, True, id="v2-rejects-v1"),
        pytest.param(
            2,
            2,
            False,
            marks=pytest.mark.xfail(
                reason="Vulnerable: minimum_auth_version=2 allows v2", strict=True
            ),
        ),
        pytest.param(3, 0, True, id="v3-rejects-v0"),
        pytest.param(3, 1, True, id="v3-rejects-v1"),
        pytest.param(3, 2, True, id="v3-rejects-v2-SECURE"),
    ],
)
async def test_replay_attack_via_version_downgrade(
    auth_pki_dir,
    functional_minion_opts,
    channel_minion_id,
    minimum_auth_version,
    attack_version,
    should_pass,
):
    """
    REGRESSION TEST: CVE-2025-62349 - Authentication Protocol Version Downgrade Attack

    This parameterized test verifies that version downgrade attacks are prevented
    based on the minimum_auth_version configuration.

    Attack Scenario:
    A malicious or compromised minion attempts to authenticate using an older protocol
    version to bypass security features introduced in newer versions (token validation,
    TTL checks, ID matching, session keys). This allows the attacker to impersonate
    other minions or maintain persistent unauthorized access.

    Test Matrix:
    - minimum_auth_version=0: VULNERABLE - accepts all versions (xfail)
    - minimum_auth_version=1: Partially secure - rejects v0 only
    - minimum_auth_version=2: Partially secure - rejects v0, v1
    - minimum_auth_version=3: SECURE - rejects v0, v1, v2 (default and recommended)

    This test uses xfail for vulnerable configurations to document the security risk.
    """
    # Create master opts with the specified minimum_auth_version
    master_opts = _create_functional_master_opts(
        auth_pki_dir, channel_minion_id, minimum_auth_version
    )

    # Setup master secrets
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "serial": multiprocessing.Value(ctypes.c_longlong, lock=False),
    }

    # Create server channel
    server_channel = salt.channel.server.ReqServerChannel.factory(master_opts)
    server_channel.auto_key = salt.daemons.masterapi.AutoKey(master_opts)
    server_channel.cache_cli = False
    server_channel.event = salt.utils.event.get_master_event(
        master_opts, master_opts["sock_dir"], listen=False
    )
    server_channel.master_key = salt.crypt.MasterKeys(master_opts)

    try:
        # Verify the configuration
        assert (
            master_opts.get("minimum_auth_version") == minimum_auth_version
        ), f"Master should have minimum_auth_version={minimum_auth_version}"

        # Read minion public key
        with salt.utils.files.fopen(
            str(auth_pki_dir / "minion_pki" / "minion.pub"), "r"
        ) as fp:
            pub_key = fp.read()

        # Create auth load for the attack version
        load = {
            "cmd": "_auth",
            "id": channel_minion_id,
            "pub": pub_key,
            "enc_algo": salt.crypt.OAEP_SHA1,
            "sig_algo": salt.crypt.PKCS1v15_SHA1,
        }

        # Add nonce for version 2+
        if attack_version >= 2:
            load["nonce"] = "test-nonce"

        # Create payload for handle_message
        payload = {"enc": "clear", "load": load, "version": attack_version}

        # Call handle_message which will enforce minimum_auth_version
        ret = await server_channel.handle_message(payload)

        # Check if attack was blocked
        if should_pass:
            # Attack should be blocked (old version rejected)
            # With proper minimum_auth_version, handle_message returns "bad load"
            assert ret == "bad load", (
                f"Expected 'bad load' for rejected version {attack_version} "
                f"with minimum {minimum_auth_version}"
            )
        else:
            # Vulnerable: attack succeeds (old version accepted)
            # Assert that auth SHOULD be rejected (but it won't be - causing xfail)
            # This assertion will FAIL for vulnerable configs, documenting the security issue
            assert ret == "bad load", (
                f"SECURITY ISSUE: Version {attack_version} was accepted "
                f"with minimum_auth_version={minimum_auth_version}"
            )

    finally:
        server_channel.close()
        if "aes" in SMaster.secrets:
            SMaster.secrets.pop("aes")


# Note: Additional functional tests for ID mismatch and token bypass
# have been covered in the unit tests. This single functional test
# demonstrates the replay attack scenario which is the most critical
# end-to-end attack vector.



"""
Runner module to directly manage the git external pillar
"""

import logging

import salt.pillar.git_pillar
import salt.utils.gitfs
from salt.exceptions import SaltRunnerError

log = logging.getLogger(__name__)


def update(branch=None, repo=None):
    """
    .. versionadded:: 2014.1.0

    .. versionchanged:: 2015.8.4
        This runner function now supports the :ref:`git_pillar
        configuration schema <git-pillar-configuration>` introduced in
        2015.8.0. Additionally, the branch and repo can now be omitted to
        update all git_pillar remotes. The return data has also changed to
        a dictionary. The values will be ``True`` only if new commits were
        fetched, and ``False`` if there were errors or no new commits were
        fetched.

    .. versionchanged:: 2018.3.0
        The return for a given git_pillar remote will now be ``None`` when no
        changes were fetched. ``False`` now is reserved only for instances in
        which there were errors.

    .. versionchanged:: 3001
        The repo parameter also matches against the repo name.

    Fetch one or all configured git_pillar remotes.

    .. note::
        This will *not* fast-forward the git_pillar cachedir on the master. All
        it does is perform a ``git fetch``. If this runner is executed with
        ``-l debug``, you may see a log message that says that the repo is
        up-to-date. Keep in mind that Salt automatically fetches git_pillar
        repos roughly every 60 seconds (or whatever
        :conf_master:`loop_interval` is set to). So, it is possible that the
        repo was fetched automatically in the time between when changes were
        pushed to the repo, and when this runner was executed. When in doubt,
        simply refresh pillar data using :py:func:`saltutil.refresh_pillar
        <salt.modules.saltutil.refresh_pillar>` and then use
        :py:func:`pillar.item <salt.modules.pillar.item>` to check if the
        pillar data has changed as expected.

    CLI Example:

    .. code-block:: bash

        # Update specific branch and repo
        salt-run git_pillar.update branch='branch' repo='https://foo.com/bar.git'
        # Update specific repo, by name
        salt-run git_pillar.update repo=myrepo
        # Update all repos
        salt-run git_pillar.update
        # Run with debug logging
        salt-run git_pillar.update -l debug
    """
    ret = {}
    for ext_pillar in __opts__.get("ext_pillar", []):
        pillar_type = next(iter(ext_pillar))
        if pillar_type != "git":
            continue
        pillar_conf = ext_pillar[pillar_type]
        pillar = salt.utils.gitfs.GitPillar(
            __opts__,
            pillar_conf,
            per_remote_overrides=salt.pillar.git_pillar.PER_REMOTE_OVERRIDES,
            per_remote_only=salt.pillar.git_pillar.PER_REMOTE_ONLY,
            global_only=salt.pillar.git_pillar.GLOBAL_ONLY,
        )
        for remote in pillar.remotes:
            # Skip this remote if it doesn't match the search criteria
            if branch is not None:
                if branch != remote.branch:
                    continue
            if repo is not None:
                if repo != remote.url and repo != getattr(remote, "name", None):
                    continue
            try:
                result = remote.fetch()
            except Exception as exc:  # pylint: disable=broad-except
                log.error(
                    "Exception '%s' caught while fetching git_pillar remote '%s'",
                    exc,
                    remote.id,
                    exc_info_on_loglevel=logging.DEBUG,
                )
                result = False
            finally:
                remote.clear_lock()
            ret[remote.id] = result

    if not ret:
        if branch is not None or repo is not None:
            raise SaltRunnerError(
                "Specified git branch/repo not found in ext_pillar config"
            )
        else:
            raise SaltRunnerError("No git_pillar remotes are configured")

    return ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import random
import string
import textwrap

import pytest

import salt.modules.cmdmod
import salt.modules.rpm_lowpkg as rpm_lowpkg
import salt.utils.path
from tests.support.mock import MagicMock, patch

try:
    import rpm

    HAS_RPM = True
except ImportError:
    HAS_RPM = False


def _called_with_root(mock):
    cmd = " ".join(mock.call_args[0][0])
    return cmd.startswith("rpm --root /")


@pytest.fixture
def configure_loader_modules():
    return {rpm_lowpkg: {"rpm": MagicMock(return_value=MagicMock)}}


def test___virtual___openeuler():
    patch_which = patch("salt.utils.path.which", return_value=True)
    with patch.dict(
        rpm_lowpkg.__grains__, {"os": "openEuler", "os_family": "openEuler"}
    ), patch_which:
        assert rpm_lowpkg.__virtual__() == "lowpkg"


def test___virtual___issabel_pbx():
    patch_which = patch("salt.utils.path.which", return_value=True)
    with patch.dict(
        rpm_lowpkg.__grains__, {"os": "Issabel Pbx", "os_family": "IssabeL PBX"}
    ), patch_which:
        assert rpm_lowpkg.__virtual__() == "lowpkg"


def test___virtual___virtuozzo():
    patch_which = patch("salt.utils.path.which", return_value=True)
    with patch.dict(
        rpm_lowpkg.__grains__, {"os": "virtuozzo", "os_family": "VirtuoZZO"}
    ), patch_which:
        assert rpm_lowpkg.__virtual__() == "lowpkg"


def test___virtual___with_no_rpm():
    patch_which = patch("salt.utils.path.which", return_value=False)
    ret = rpm_lowpkg.__virtual__()
    assert isinstance(ret, tuple)
    assert ret[0] is False


# 'list_pkgs' function tests: 2


def test_list_pkgs():
    """
    Test if it list the packages currently installed in a dict
    """
    mock = MagicMock(return_value="")
    with patch.dict(rpm_lowpkg.__salt__, {"cmd.run": mock}):
        assert rpm_lowpkg.list_pkgs() == {}
        assert not _called_with_root(mock)


def test_list_pkgs_root():
    """
    Test if it list the packages currently installed in a dict,
    called with root parameter
    """
    mock = MagicMock(return_value="")
    with patch.dict(rpm_lowpkg.__salt__, {"cmd.run": mock}):
        rpm_lowpkg.list_pkgs(root="/")
        assert _called_with_root(mock)


# 'verify' function tests: 2


def test_verify():
    """
    Test if it runs an rpm -Va on a system, and returns the
    results in a dict
    """
    mock = MagicMock(
        return_value={"stdout": "", "stderr": "", "retcode": 0, "pid": 12345}
    )
    with patch.dict(rpm_lowpkg.__salt__, {"cmd.run_all": mock}):
        assert rpm_lowpkg.verify("httpd") == {}
        assert not _called_with_root(mock)


def test_verify_root():
    """
    Test if it runs an rpm -Va on a system, and returns the
    results in a dict, called with root parameter
    """
    mock = MagicMock(
        return_value={"stdout": "", "stderr": "", "retcode": 0, "pid": 12345}
    )
    with patch.dict(rpm_lowpkg.__salt__, {"cmd.run_all": mock}):
        rpm_lowpkg.verify("httpd", root="/")
        assert _called_with_root(mock)


# 'file_list' function tests: 2


def test_file_list():
    """
    Test if it list the files that belong to a package.
    """
    mock = MagicMock(return_value="")
    with patch.dict(rpm_lowpkg.__salt__, {"cmd.run": mock}):
        assert rpm_lowpkg.file_list("httpd") == {"errors": [], "files": []}
        assert not _called_with_root(mock)


def test_file_list_root():
    """
    Test if it list the files that belong to a package, using the
    root parameter.
    """

    mock = MagicMock(return_value="")
    with patch.dict(rpm_lowpkg.__salt__, {"cmd.run": mock}):
        rpm_lowpkg.file_list("httpd", root="/")
        assert _called_with_root(mock)


# 'file_dict' function tests: 2


def test_file_dict():
    """
    Test if it list the files that belong to a package
    """
    mock = MagicMock(return_value="")
    with patch.dict(rpm_lowpkg.__salt__, {"cmd.run": mock}):
        assert rpm_lowpkg.file_dict("httpd") == {"errors": [], "packages": {}}
        assert not _called_with_root(mock)


def test_file_dict_root():
    """
    Test if it list the files that belong to a package
    """
    mock = MagicMock(return_value="")
    with patch.dict(rpm_lowpkg.__salt__, {"cmd.run": mock}):
        rpm_lowpkg.file_dict("httpd", root="/")
        assert _called_with_root(mock)


# 'owner' function tests: 1


def test_owner():
    """
    Test if it return the name of the package that owns the file.
    """
    assert rpm_lowpkg.owner() == ""

    ret = "file /usr/bin/salt-jenkins-build is not owned by any package"
    mock = MagicMock(return_value=ret)
    with patch.dict(rpm_lowpkg.__salt__, {"cmd.run_stdout": mock}):
        assert rpm_lowpkg.owner("/usr/bin/salt-jenkins-build") == ""
        assert not _called_with_root(mock)

    ret = {
        "/usr/bin/vim": "vim-enhanced-7.4.160-1.e17.x86_64",
        "/usr/bin/python": "python-2.7.5-16.e17.x86_64",
    }
    mock = MagicMock(
        side_effect=[
            "python-2.7.5-16.e17.x86_64",
            "vim-enhanced-7.4.160-1.e17.x86_64",
        ]
    )
    with patch.dict(rpm_lowpkg.__salt__, {"cmd.run_stdout": mock}):
        assert rpm_lowpkg.owner("/usr/bin/python", "/usr/bin/vim") == ret
        assert not _called_with_root(mock)


def test_owner_root():
    """
    Test if it return the name of the package that owns the file,
    using the parameter root.
    """
    assert rpm_lowpkg.owner() == ""

    ret = "file /usr/bin/salt-jenkins-build is not owned by any package"
    mock = MagicMock(return_value=ret)
    with patch.dict(rpm_lowpkg.__salt__, {"cmd.run_stdout": mock}):
        rpm_lowpkg.owner("/usr/bin/salt-jenkins-build", root="/")
        assert _called_with_root(mock)


# 'checksum' function tests: 2


def test_checksum():
    """
    Test if checksum validate as expected
    """
    ret = {
        "file1.rpm": True,
        "file2.rpm": False,
        "file3.rpm": False,
    }

    mock = MagicMock(side_effect=[True, 0, True, 1, False, 0])
    with patch.dict(
        rpm_lowpkg.__salt__, {"file.file_exists": mock, "cmd.retcode": mock}
    ):
        assert rpm_lowpkg.checksum("file1.rpm", "file2.rpm", "file3.rpm") == ret
        assert not _called_with_root(mock)


def test_checksum_root():
    """
    Test if checksum validate as expected, using the parameter
    root
    """
    mock = MagicMock(side_effect=[True, 0])
    with patch.dict(
        rpm_lowpkg.__salt__, {"file.file_exists": mock, "cmd.retcode": mock}
    ):
        rpm_lowpkg.checksum("file1.rpm", root="/")
        assert _called_with_root(mock)


@pytest.mark.skipif(not HAS_RPM, reason="python rpm module not available")
def test_version_cmp_rpm_lib():
    """
    Test package version when each library is installed
    """
    patch_cmd = patch.dict(
        rpm_lowpkg.__salt__, {"cmd.run_all": salt.modules.cmdmod.run_all}
    )
    patch_rpm = patch("salt.modules.rpm_lowpkg.HAS_RPM", True)

    with patch_rpm, patch_cmd:
        assert rpm_lowpkg.version_cmp("1", "2") == -1
        assert rpm_lowpkg.version_cmp("2.9.1-6.el7_2.3", "2.9.1-6.el7.4") == -1
        assert rpm_lowpkg.version_cmp("3.2", "3.0") == 1
        assert rpm_lowpkg.version_cmp("3.0", "3.0") == 0
        assert rpm_lowpkg.version_cmp("1:2.9.1-6.el7_2.3", "2.9.1-6.el7.4") == 1
        assert rpm_lowpkg.version_cmp("1:2.9.1-6.el7_2.3", "1:2.9.1-6.el7.4") == -1
        assert rpm_lowpkg.version_cmp("2:2.9.1-6.el7_2.3", "1:2.9.1-6.el7.4") == 1
        assert rpm_lowpkg.version_cmp("3:2.9.1-6.el7.4", "3:2.9.1-6.el7.4") == 0
        assert rpm_lowpkg.version_cmp("3:2.9.1-6.el7.4", "3:2.9.1-7.el7.4") == -1
        assert rpm_lowpkg.version_cmp("3:2.9.1-8.el7.4", "3:2.9.1-7.el7.4") == 1
        assert rpm_lowpkg.version_cmp("3.23-6.el9", "3.23") == 0
        assert rpm_lowpkg.version_cmp("3.23", "3.23-6.el9") == 0
        assert (
            rpm_lowpkg.version_cmp("release_web_294-6", "release_web_294_applepay-1")
            == -1
        )


def test_version_cmp_rpm():
    """
    Test package version if RPM-Python is installed

    :return:
    """
    mock_label = MagicMock(return_value=-1)
    mock_log = MagicMock()
    patch_label = patch("salt.modules.rpm_lowpkg.rpm.labelCompare", mock_label)
    patch_log = patch("salt.modules.rpm_lowpkg.log", mock_log)
    patch_rpm = patch("salt.modules.rpm_lowpkg.HAS_RPM", True)
    with patch_label, patch_rpm, patch_log:
        assert -1 == rpm_lowpkg.version_cmp("1", "2")
        assert not mock_log.warning.called
        assert mock_label.called


def test_version_cmp_python():
    """
    Test package version if falling back to python

    :return:
    """
    mock_log = MagicMock()
    patch_rpm = patch("salt.modules.rpm_lowpkg.HAS_RPM", False)
    patch_log = patch("salt.modules.rpm_lowpkg.log", mock_log)

    with patch_rpm, patch_log:
        assert -1 == rpm_lowpkg.version_cmp("1", "2")
        assert not mock_log.warning.called
        assert mock_log.info.called
        assert (
            mock_log.info.mock_calls[0][1][0]
            == "Install a package that provides rpm.labelCompare for faster version comparisons."
        )


def _parse_label(label: str):
    """Split full label into (epoch, version, release) for rpm.labelCompare."""
    epoch = None
    version = None
    release = None
    if ":" in label:
        epoch, rest = label.split(":", 1)
    else:
        rest = label
    if "-" in rest:
        version, release = rest.split("-", 1)
    else:
        version = rest
    return (epoch, version, release)


VERSION_CASES = [
    # Basic equality and ordering
    ("1", "1", 0),
    ("1", "2", -1),
    ("2", "1", 1),
    ("1.0", "1.0", 0),
    ("1.0", "2.0", -1),
    ("2.0", "1.0", 1),
    ("2.0.1", "2.0.1", 0),
    ("2.0", "2.0.1", -1),
    ("2.0.1", "2.0", 1),
    # Epoch precedence
    ("0:1.0-1", "1.0-1", 0),
    ("1:1.0-1", "0:9.9-9", 1),
    ("0:9.9-9", "1:1.0-1", -1),
    ("02:1.0-1", "2:1.0-1", 0),
    # Version vs release precedence (version decides before release)
    ("1.0.1-1", "1.0-9", 1),
    ("1.0-9", "1.0.1-1", -1),
    # Numeric vs numeric
    ("2.0-1", "10.0-1", -1),
    ("10.0-1", "2.0-1", 1),
    ("2.10-1", "2.2-1", 1),
    ("2.02-1", "2.2-1", 0),
    ("10.0001", "10.0001", 0),
    ("10.0001", "10.1", 0),
    ("10.1", "10.0001", 0),
    ("10.0001", "10.0039", -1),
    ("10.0039", "10.0001", 1),
    ("010.0039", "10.0039", 0),
    ("4.999.9", "5.0", -1),
    ("5.0", "4.999.9", 1),
    # Long numeric segments
    ("2.12345678901234567890-1", "2.12345678901234567891-1", -1),
    # Date-like numbers
    ("20101121", "20101121", 0),
    ("20101121", "20101122", -1),
    ("20101122", "20101121", 1),
    # 'p' in numeric-ish segments (treated as alphabetic break)
    ("5.5p1", "5.5p1", 0),
    ("5.5p1", "5.5p2", -1),
    ("5.5p2", "5.5p1", 1),
    ("5.5p10", "5.5p10", 0),
    ("5.5p1", "5.5p10", -1),
    ("5.5p10", "5.5p1", 1),
    ("5.5p2", "5.6p1", -1),
    ("5.6p1", "5.5p2", 1),
    ("5.6p1", "6.5p1", -1),
    ("6.5p1", "5.6p1", 1),
    # Dot before alpha: dots split segments but alpha still compares after
    ("2.0.1a", "2.0.1a", 0),
    ("2.0.1a", "2.0.1", 1),
    ("2.0.1", "2.0.1a", -1),
    ("6.0.rc1", "6.0", 1),
    ("6.0", "6.0.rc1", -1),
    # other alphanumeric
    ("10xyz", "10.1xyz", -1),
    ("10.1xyz", "10xyz", 1),
    ("xyz10", "xyz10", 0),
    ("xyz10", "xyz10.1", -1),
    ("xyz10.1", "xyz10", 1),
    # capital letters
    ("A", "A", 0),
    ("A", "B", -1),
    ("B", "a", -1),
    ("A", "a", -1),
    ("Ab", "Ab", 0),
    ("A0", "A0", 0),
    ("A", "a", -1),
    ("a", "A", 1),
    ("Ab", "aB", -1),
    ("A.1", "a.1", -1),
    ("Abc1", "abc1", -1),
    ("abc1", "Abc1", 1),
    ("A1b2", "A1B2", 1),
    ("1A", "1a", -1),
    ("1.A", "1.a", -1),
    ("1.A.2", "1.a.2", -1),
    ("1.0A", "1.0a", -1),
    ("1.0a", "1.0A", 1),
    ("1.0Aalpha", "1.0aalpha", -1),
    # Alphanumeric segment vs pure-numeric segment ordering
    # Alphanumeric segments (start with letters) should sort after pure-numeric versions;
    # dotted alpha+numeric combos compare as expected and equality is preserved.
    ("xyz.4", "xyz.4", 0),
    ("xyz.4", "8", -1),
    ("8", "xyz.4", 1),
    ("xyz.4", "2", -1),
    ("2", "xyz.4", 1),
    # Alphabetic tails and mixed alpha-numeric comparisons
    # Alphabetic suffixes are compared lexicographically; when alpha tails share a prefix
    # the shorter tail sorts earlier. Mixed numeric+alpha+numeric tokens use the numeric
    # prefix first, then alphabetic ordering to break ties.
    ("a", "a", 0),
    ("2a-1", "2-1", 1),
    ("2-1", "2a-1", -1),
    ("2alpha-1", "2beta-1", -1),
    ("2.0.1a", "2.0.1a", 0),
    ("alpha-1", "beta-1", -1),
    ("beta-1", "alpha-1", 1),
    ("10b2", "10a1", 1),
    ("10a2", "10b2", -1),
    ("1.0aa", "1.0aa", 0),
    ("1.0a", "1.0aa", -1),
    ("1.0aa", "1.0a", 1),
    ("1.0~A", "1.0~a", -1),
    ("1.0~a", "1.0~A", 1),
    ("1.0^A", "1.0^a", -1),
    ("1.0^a", "1.0^A", 1),
    ("1.0+A", "1.0+a", -1),
    ("1.0+a", "1.0+A", 1),
    # Tilde: pre-release; sorts older than anything without it
    ("2.0~beta-1", "2.0-1", -1),
    ("2.0-1", "2.0~beta-1", 1),
    ("2.0~beta-1", "2.0~rc-1", -1),  # beta < rc lexicographically
    ("1.0~rc1", "1.0~rc1", 0),
    ("1.0~rc1", "1.0", -1),
    ("1.0", "1.0~rc1", 1),
    ("1.0~rc1", "1.0~rc2", -1),
    ("1.0~rc2", "1.0~rc1", 1),
    # Tilde chaining
    ("1.0~rc1~git123", "1.0~rc1~git123", 0),
    ("1.0~rc1~git123", "1.0~rc1", -1),
    ("1.0~rc1", "1.0~rc1~git123", 1),
    # Caret: post-release; newer than base but lower than next increment
    ("1.0^", "1.0^", 0),
    ("1.0^", "1.0", 1),
    ("1.0", "1.0^", -1),
    ("1.0^git1", "1.0", 1),
    ("1.0", "1.0^git1", -1),
    ("1.0^git1", "1.0^git1", 0),
    ("1.0^git1", "1.0^git2", -1),
    ("1.0^git2", "1.0^git1", 1),
    ("1.0^git9", "1.0.1", -1),  # caret block still less than next numeric bump
    ("1.0.1", "1.0^git9", 1),
    ("1.0^git1", "1.01", -1),
    ("1.01", "1.0^git1", 1),
    ("1.0^20160101", "1.0^20160101", 0),
    ("1.0^20160101", "1.0.1", -1),
    ("1.0.1", "1.0^20160101", 1),
    ("1.0^20160101^git1", "1.0^20160101^git1", 0),
    ("1.0^20160102", "1.0^20160101^git1", 1),
    ("1.0^20160101^git1", "1.0^20160102", -1),
    # Caret + tilde
    ("1.0~rc1^git1", "1.0~rc1", 1),
    ("1.0~rc1", "1.0~rc1^git1", -1),
    ("1.0^git1~pre", "1.0^git1", -1),
    ("1.0^git1", "1.0^git1~pre", 1),
    # Separators: '.', '_', '+' are equivalent and collapse
    ("2_0", "2.0", 0),
    ("2.0", "2_0", 0),
    ("2+0", "2.0", 0),
    ("2.0", "2+0", 0),
    ("2_0", "2+0", 0),
    # Collapsing multiple separators
    ("1+.+0", "1.0", 0),
    ("1_._0", "1.0", 0),
    ("1+_+0", "1.0", 0),
    # Plus/underscore equivalences at segment edges
    ("a+", "a_", 0),
    ("a+", "a+", 0),
    ("a+", "a_", 0),
    ("a_", "a+", 0),
    ("+a", "+a", 0),
    ("+a", "_a", 0),
    ("_a", "+a", 0),
    ("+_", "+_", 0),
    ("+_", "_+", 0),
    ("+", "_", 0),
    ("_", "+", 0),
    # Mixed separators with numbers
    ("1+2", "1.2", 0),
    ("1_2", "1.2", 0),
    ("1+_2", "1.2", 0),
    # Release segment variations
    ("2.0-1-alpha", "2.0-1-beta", -1),
    ("2.0-1-beta", "2.0-1-alpha", 1),
    ("2.0-1-alpha-1", "2.0-1-alpha-2", -1),
    ("2.0-1.alpha-beta.1", "2.0-1.alpha-beta.2", -1),
    ("2.0-1.alpha-beta.2", "2.0-1.alpha-beta.1", 1),
    ("2.0-1.alpha-beta.gamma", "2.0-1.alpha-beta.delta", 1),
    # Release containing periods + trailing zeros (RPM: longer wins, even if .0)
    ("2.0-1.0.0", "2.0-1.0", 1),
    ("2.0-1.0.1", "2.0-1.0", 1),
    ("2.0-1.alpha.1", "2.0-1.alpha.2", -1),
    ("2.0-1.alpha.01", "2.0-1.alpha.1", 0),
    # Empty release equivalence and ordering
    ("2.0", "2.0-", 0),
    ("2.0-", "2.0-1", 0),
    # Different number of segments
    ("1.0", "1.0.1", -1),
    ("1.0.1", "1.0", 1),
    ("1.0", "1.0.0", -1),
    ("1.0.0.1", "1.0", 1),
    ("1.0a", "1.0.a", 0),
    # Different number of release segments
    ("1.0-1", "1.0-1.1", -1),
    ("1.0-1.1", "1.0-1", 1),
    ("1.0-1", "1.0-1.0", -1),
    ("1.0-1.0.1", "1.0-1", 1),
    # Mixed alpha/numeric with extra segments in version
    ("1.0alpha", "1.0alpha.1", -1),
    ("1.0alpha.1", "1.0alpha", 1),
    ("1.0alpha.0", "1.0alpha", 1),
    # Mixed alpha/numeric with extra segments in release
    ("1.0-1-alpha", "1.0-1-alpha.1", -1),
    ("1.0-1-alpha.1", "1.0-1-alpha", 1),
    ("1.0-1-alpha.0", "1.0-1-alpha", 1),
    # Multiple hyphens inside the release
    ("1.0-1-alpha-beta", "1.0-1-alpha-gamma", -1),
    ("1.0-1-alpha-gamma", "1.0-1-alpha-beta", 1),
    ("1.0-1-alpha-beta", "1.0-1-alpha", 1),
    ("1.0-1-alpha", "1.0-1-alpha-beta", -1),
    ("1.0-1-alpha-1", "1.0-1-alpha", 1),
    ("1.0-1-alpha", "1.0-1-alpha-1", -1),
    ("2.0-1-alpha-beta.2", "2.0-1-alpha-beta.10", -1),
    ("2.0-1-alpha-beta.10", "2.0-1-alpha-beta.2", 1),
    # Releases containing multiple numeric hyphen segments
    ("1.0-1-1", "1.0-1-2", -1),
    ("1.0-1-2", "1.0-1-1", 1),
    ("1.0-1-01", "1.0-1-1", 0),
    # Releases with separators and multiple operator segments
    ("1.0-1.alpha-beta-1", "1.0-1.alpha-beta-2", -1),
    ("1.0-1.alpha-beta-2", "1.0-1.alpha-beta-1", 1),
    # Explicit empty vs multiple-hyphen release forms
    ("1.0-", "1.0-1-alpha", 0),
    ("1.0-1-alpha", "1.0-", 0),
    # Longer release wins when numeric/alpha tie in earlier segments
    ("3.0-1.0.0-0", "3.0-1.0.0", 1),
    ("3.0-1.0.0", "3.0-1.0.0-0", -1),
    # Same version, one side has no release -> treat as equal
    ("1.0-1", "1.0", 0),
    ("1.0", "1.0-1", 0),
    ("2.3.4-5", "2.3.4", 0),
    ("2.3.4", "2.3.4-5", 0),
    # Different versions, release ignored if one side missing -> version decides
    ("1.0-2", "1.1", -1),
    ("1.1", "1.0-2", 1),
    ("2.0-3", "2.0.1", -1),
    ("2.0.1", "2.0-3", 1),
    # Epoch differences still take precedence even when release missing
    ("1:1.0-1", "1.0", 1),
    ("1.0", "1:1.0-1", -1),
    ("0:2.0-1", "1:1.9", -1),
    ("1:1.9", "0:2.0-1", 1),
    # Both sides have version segments that compare alphabetically; release ignored when missing
    ("1.0a-1", "1.0a", 0),
    ("1.0a", "1.0a-2", 0),
    ("1.0~rc1-1", "1.0~rc1", 0),
    ("1.0~rc1", "1.0~rc1-1", 0),
    # Operator blocks: caret/tilde interactions, release ignored when missing
    ("1.0^git1-1", "1.0^git1", 0),
    ("1.0^git1", "1.0^git1-2", 0),
    (
        "1.0~beta-1",
        "1.0",
        -1,
    ),  # tilde makes version older than base even if release present on left
    ("1.0", "1.0~beta-1", 1),
    # One side has complex release, other has no release; version decides when different
    ("2.0.1-10.alpha", "2.0.1", 0),
    ("2.0.1", "2.0.1-10.alpha", 0),
    ("2.0.2-1", "2.0.10", -1),
    ("2.0.10", "2.0.2-1", 1),
    # Special chars in wild
    ("2.0_git20210101-1", "2.0_git20201231-1", 1),
    ("2.0+dfsg-1", "2.0-1", 1),
    # More separator/operator chains
    ("1.0+rc1", "1.0_rc1", 0),
    ("1.0+rc1", "1.0.rc1", 0),
    ("1.0+rc1^git1", "1.0.rc1^git1", 0),
    ("1.0~beta+1", "1.0~beta.1", 0),
    ("1.0~beta_1", "1.0~beta.1", 0),
    # Caret blocks vs longer alpha tail
    ("1.0^a", "1.0a", -1),
    ("1.0a", "1.0^a", 1),
    # Tilde vs caret at same position
    ("1.0~a", "1.0^a", -1),  # tilde is always older
    ("1.0^a", "1.0~a", 1),
    # Chained: tilde inside caret block
    ("1.0^a~b", "1.0^a", -1),
    ("1.0^a", "1.0^a~b", 1),
    # Chained: caret after tilde block
    ("1.0~a^b", "1.0~a", 1),
    ("1.0~a", "1.0~a^b", -1),
    # Multi-caret sequence
    ("1.0^a^b", "1.0^a^c", -1),
    ("1.0^a^c", "1.0^a^b", 1),
    # Mixed separators around operators
    ("1.0+^git1", "1.0^git1", 0),
    ("1.0_^git1", "1.0^git1", 0),
    ("1.0+~rc1", "1.0~rc1", 0),
    ("1.0_~rc1", "1.0~rc1", 0),
]


@pytest.mark.parametrize("label1, label2, expected", VERSION_CASES)
def test_version_cmp_expected(label1, label2, expected):
    patch_rpm = patch("salt.modules.rpm_lowpkg.HAS_RPM", False)
    with patch_rpm:
        result = rpm_lowpkg.version_cmp(label1, label2)
        assert (
            result == expected
        ), f"{label1} vs {label2} => {result}, expected {expected}"


@pytest.mark.parametrize("label1, label2, _", VERSION_CASES)
@pytest.mark.skipif(not HAS_RPM, reason="python rpm module not available")
def test_version_cmp_matches_rpm(label1, label2, _):
    evr1 = _parse_label(label1)
    evr2 = _parse_label(label2)
    py_result = rpm_lowpkg.version_cmp(label1, label2)
    rpm_result = rpm.labelCompare(evr1, evr2)
    assert py_result == rpm_result, (
        f"Mismatch for {label1} vs {label2}: "
        f"pure-Python={py_result}, rpm={rpm_result}"
    )


def test_symmetry_and_transitivity():
    # Small chain that exercises epoch, version, and release ordering properties
    chain = [
        "0:1.0-1",
        "0:1.0-2",
        "0:1.0^git1-1",
        "0:1.0.1-1",
        "1:0.1-1",
    ]
    # Check reflexivity and antisymmetry
    for a in chain:
        assert rpm_lowpkg.version_cmp(a, a) == 0
    # Check ordering across the chain is strictly increasing
    for i in range(len(chain) - 1):
        a, b = chain[i], chain[i + 1]
        assert rpm_lowpkg.version_cmp(a, b) < 0
        assert rpm_lowpkg.version_cmp(b, a) > 0


# ===== Optional fuzzing =====
@pytest.mark.skipif(not HAS_RPM, reason="python rpm module not available")
def test_fuzz_against_rpm():
    """
    Deterministic fuzz over a restricted alphabet to spot regressions against rpm.labelCompare().
    """
    rng = random.Random(1337)
    alpha = string.ascii_lowercase
    digits = string.digits
    seps = "._+"
    ops = "~^"

    # Limit length to keep the space sane for CI
    def rand_token():
        t = []
        for _ in range(rng.randint(1, 6)):
            choice = rng.random()
            if choice < 0.55:
                t.append(rng.choice(digits))
            elif choice < 0.85:
                t.append(rng.choice(alpha))
            elif choice < 0.95:
                t.append(rng.choice(seps))
            else:
                t.append(rng.choice(ops))
        return "".join(t).strip("._+")

    def rand_evr():
        # 30% chance epoch, 70% none
        epoch = str(rng.randint(0, 3)) if rng.random() < 0.3 else None
        version = rand_token() or "1"
        release = rand_token() if rng.random() < 0.6 else None
        return (epoch, version, release)

    for _ in range(2000):
        evr1 = rand_evr()
        evr2 = rand_evr()
        l1 = (
            ("" if evr1[0] is None else f"{evr1[0]}:")
            + evr1[1]
            + ("" if evr1[2] is None else f"-{evr1[2]}")
        )
        l2 = (
            ("" if evr2[0] is None else f"{evr2[0]}:")
            + evr2[1]
            + ("" if evr2[2] is None else f"-{evr2[2]}")
        )
        py = rpm_lowpkg.version_cmp(l1, l2)
        rp = rpm.labelCompare(evr1, evr2)
        if py < 0:
            assert rp < 0, (l1, l2, evr1, evr2, py, rp)
        elif py > 0:
            assert rp > 0, (l1, l2, evr1, evr2, py, rp)
        else:
            assert rp == 0, (l1, l2, evr1, evr2, py, rp)


@pytest.mark.skip_on_windows
def test_info():
    """
    Confirm that a nonzero retcode does not raise an exception.
    """
    rpm_out = textwrap.dedent(
        """\
        name: bash
        relocations: (not relocatable)
        version: 4.4.19
        vendor: CentOS
        release: 10.el8
        build_date_time_t: 1573230816
        build_date: 1573230816
        install_date_time_t: 1578952147
        install_date: 1578952147
        build_host: x86-01.mbox.centos.org
        group: Unspecified
        source_rpm: bash-4.4.19-10.el8.src.rpm
        size: 6930068
        arch: x86_64
        license: GPLv3+
        signature: RSA/SHA256, Wed Dec  4 22:45:04 2019, Key ID 05b555b38483c65d
        packager: CentOS Buildsys <bugs@centos.org>
        url: https://www.gnu.org/software/bash
        summary: The GNU Bourne Again shell
        edition: 4.4.19-10.el8
        description:
        The GNU Bourne Again shell (Bash) is a shell or command language
        interpreter that is compatible with the Bourne shell (sh). Bash
        incorporates useful features from the Korn shell (ksh) and the C shell
        (csh). Most sh scripts can be run by bash without modification.
        -----"""
    )
    dunder_salt = {
        "cmd.run_stdout": MagicMock(return_value="LONGSIZE"),
        "cmd.run_all": MagicMock(
            return_value={
                "retcode": 123,
                "stdout": rpm_out,
                "stderr": "",
                "pid": 12345,
            }
        ),
    }
    expected = {
        "bash": {
            "relocations": "(not relocatable)",
            "version": "4.4.19",
            "vendor": "CentOS",
            "release": "10.el8",
            "build_date_time_t": 1573230816,
            "build_date": "2019-11-08T16:33:36Z",
            "install_date_time_t": 1578952147,
            "install_date": "2020-01-13T21:49:07Z",
            "build_host": "x86-01.mbox.centos.org",
            "group": "Unspecified",
            "source_rpm": "bash-4.4.19-10.el8.src.rpm",
            "size": "6930068",
            "arch": "x86_64",
            "license": "GPLv3+",
            "signature": "RSA/SHA256, Wed Dec  4 22:45:04 2019, Key ID 05b555b38483c65d",
            "packager": "CentOS Buildsys <bugs@centos.org>",
            "url": "https://www.gnu.org/software/bash",
            "summary": "The GNU Bourne Again shell",
            "description": "The GNU Bourne Again shell (Bash) is a shell or command language\ninterpreter that is compatible with the Bourne shell (sh). Bash\nincorporates useful features from the Korn shell (ksh) and the C shell\n(csh). Most sh scripts can be run by bash without modification.",
        }
    }
    with patch.dict(rpm_lowpkg.__salt__, dunder_salt):
        result = rpm_lowpkg.info("bash")
        assert result == expected, result



import pytest

import salt.utils.path
import salt.utils.platform
from tests.support.case import ModuleCase
from tests.support.sminion import create_sminion


@pytest.mark.skip_unless_on_windows
@pytest.mark.windows_whitelisted
@pytest.mark.destructive_test
@pytest.mark.slow_test
class ChocolateyModuleTest(ModuleCase):
    """
    Validate Chocolatey module
    """

    @classmethod
    def setUpClass(cls):
        """
        Ensure that Chocolatey is installed
        """
        if salt.utils.path.which("chocolatey.exe") is None:
            sminion = create_sminion()
            sminion.functions.chocolatey.bootstrap()

    def test_list_(self):
        ret = self.run_function("chocolatey.list", narrow="adobereader", exact=True)
        self.assertTrue("adobereader" in ret)

    def test_list_sources(self):
        ret = self.run_function("chocolatey.list_sources")
        self.assertTrue("chocolatey" in ret.keys())



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    TestCase for salt.modules.serverdensity_device
"""

import pytest

import salt.modules.serverdensity_device as serverdensity_device
import salt.utils.json
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


class MockRequests:
    """
    Mock smtplib class
    """

    flag = None
    content = """{"message": "Invalid token", "errors": [{"type": "invalid_token", "subject": "token"}]}"""
    status_code = None

    def __init__(self):
        self.url = None
        self.data = None
        self.kwargs = None

    def return_request(self, url, data=None, **kwargs):
        """
        Mock request method.
        """
        self.url = url
        self.data = data
        self.kwargs = kwargs
        requests = MockRequests()
        if self.flag == 1:
            requests.status_code = 401
        else:
            requests.status_code = 200
        return requests

    def post(self, url, data=None, **kwargs):
        """
        Mock post method.
        """
        return self.return_request(url, data, **kwargs)

    def delete(self, url, **kwargs):
        """
        Mock delete method.
        """
        return self.return_request(url, **kwargs)

    def get(self, url, **kwargs):
        """
        Mock get method.
        """
        return self.return_request(url, **kwargs)

    def put(self, url, data=None, **kwargs):
        """
        Mock put method.
        """
        return self.return_request(url, data, **kwargs)


@pytest.fixture
def configure_loader_modules():
    return {serverdensity_device: {"requests": MockRequests()}}


@pytest.fixture
def mock_json_loads():
    return MagicMock(side_effect=ValueError())


def test_get_sd_auth():
    """
    Tests if it returns requested Server Density
    authentication value from pillar.
    """
    with patch.dict(serverdensity_device.__pillar__, {"serverdensity": False}):
        pytest.raises(CommandExecutionError, serverdensity_device.get_sd_auth, "1")

    with patch.dict(serverdensity_device.__pillar__, {"serverdensity": {"1": "salt"}}):
        assert serverdensity_device.get_sd_auth("1") == "salt"

        pytest.raises(CommandExecutionError, serverdensity_device.get_sd_auth, "2")


def test_create(mock_json_loads):
    """
    Tests if it create device in Server Density.
    """
    with patch.dict(
        serverdensity_device.__pillar__, {"serverdensity": {"api_token": "salt"}}
    ):
        assert serverdensity_device.create("rich_lama", group="lama_band")

        with patch.object(salt.utils.json, "loads", mock_json_loads):
            pytest.raises(
                CommandExecutionError,
                serverdensity_device.create,
                "rich_lama",
                group="lama_band",
            )

        MockRequests.flag = 1
        assert serverdensity_device.create("rich_lama", group="lama_band") is None


def test_delete(mock_json_loads):
    """
    Tests if it delete a device from Server Density.
    """
    with patch.dict(
        serverdensity_device.__pillar__, {"serverdensity": {"api_token": "salt"}}
    ):
        MockRequests.flag = 0
        assert serverdensity_device.delete("51f7eaf")

        with patch.object(salt.utils.json, "loads", mock_json_loads):
            pytest.raises(CommandExecutionError, serverdensity_device.delete, "51f7eaf")

        MockRequests.flag = 1
        assert serverdensity_device.delete("51f7eaf") is None


def test_ls(mock_json_loads):
    """
    Tests if it list devices in Server Density.
    """
    with patch.dict(
        serverdensity_device.__pillar__, {"serverdensity": {"api_token": "salt"}}
    ):
        MockRequests.flag = 0
        assert serverdensity_device.ls(name="lama")

        with patch.object(salt.utils.json, "loads", mock_json_loads):
            pytest.raises(CommandExecutionError, serverdensity_device.ls, name="lama")

        MockRequests.flag = 1
        assert serverdensity_device.ls(name="lama") is None


def test_update(mock_json_loads):
    """
    Tests if it updates device information in Server Density.
    """
    with patch.dict(
        serverdensity_device.__pillar__, {"serverdensity": {"api_token": "salt"}}
    ):
        MockRequests.flag = 0
        assert serverdensity_device.update("51f7eaf", name="lama")

        with patch.object(salt.utils.json, "loads", mock_json_loads):
            pytest.raises(
                CommandExecutionError,
                serverdensity_device.update,
                "51f7eaf",
                name="lama",
            )

        MockRequests.flag = 1
        assert serverdensity_device.update("51f7eaf", name="lama") is None


def test_install_agent():
    """
    Tests if it downloads Server Density installation agent,
    and installs sd-agent with agent_key.
    """
    mock = MagicMock(return_value=True)
    with patch.dict(
        serverdensity_device.__pillar__, {"serverdensity": {"account_url": "salt"}}
    ):
        with patch.dict(serverdensity_device.__salt__, {"cmd.run": mock}):
            with patch.dict(serverdensity_device.__opts__, {"cachedir": "/"}):
                assert serverdensity_device.install_agent("51f7e")


def test_install_agent_v2():
    """
    Tests if it downloads Server Density installation agent,
    and installs sd-agent with agent_key.
    """
    mock = MagicMock(return_value=True)
    with patch.dict(
        serverdensity_device.__pillar__, {"serverdensity": {"account_name": "salt"}}
    ):
        with patch.dict(serverdensity_device.__salt__, {"cmd.run": mock}):
            with patch.dict(serverdensity_device.__opts__, {"cachedir": "/"}):
                assert serverdensity_device.install_agent("51f7e", agent_version=2)



from salt.utils.aggregation import Map, Scalar, aggregate
from tests.support.unit import TestCase


class TestAggregation(TestCase):
    def test_merging(self):
        a = {"foo": 42, "bar": "first"}
        b = {"bar": "second"}

        c, d = "first", "second"

        # introspection
        for level in (None, False, 0, []):
            assert aggregate(a, b, level=level) == {"bar": "second"}
            assert aggregate(c, d, level=level) == "second"

        # first level aggregation
        for level in (1, [1, 0], [True], "10000"):
            assert aggregate(a, b, level=level) == {"foo": 42, "bar": "second"}
            assert aggregate(c, d, level=level) == ["first", "second"]

        # 1-2nd level aggregation
        for level in (2, [1, 1], [True, True], "11"):
            assert aggregate(a, b, level=level) == {
                "foo": 42,
                "bar": ["first", "second"],
            }, aggregate(a, b, level=2)
            assert aggregate(c, d, level=level) == ["first", "second"]

        # full aggregation
        for level in (True,):
            assert aggregate(a, b, level=level) == {
                "foo": 42,
                "bar": ["first", "second"],
            }
            assert aggregate(c, d, level=level) == ["first", "second"]

    def test_nested(self):
        a = {"foo": {"bar": "first"}}
        b = {"foo": {"bar": "second"}}
        assert aggregate(a, b) == {"foo": {"bar": "second"}}, aggregate(a, b)

        a = {"foo": {"bar": Scalar("first")}}
        b = {"foo": {"bar": Scalar("second")}}

        assert aggregate(a, b) == {"foo": {"bar": ["first", "second"]}}, aggregate(a, b)

    def test_introspection(self):

        a = {"foo": {"lvl1": {"lvl2-a": "first", "lvl2-b": "first"}}}

        b = {"foo": {"lvl1": {"lvl2-a": "second"}}}

        assert aggregate(a, b) == {"foo": {"lvl1": {"lvl2-a": "second"}}}, aggregate(
            a, b
        )

    def test_instruction(self):
        a = {"foo": Map({"bar": Scalar("first")})}
        b = {"foo": Map({"bar": Scalar("second")})}
        c = {"foo": Map({"another": "value"})}
        result = aggregate(c, aggregate(a, b), level=2)
        assert result == {
            "foo": {"bar": ["first", "second"], "another": "value"}
        }, result



import os

import pytest

import salt.config
from tests.conftest import FIPS_TESTRUN


@pytest.fixture
def minion_opts(tmp_path):
    """
    Default minion configuration with relative temporary paths to not require root permissions.
    """
    root_dir = tmp_path / "minion"
    opts = salt.config.DEFAULT_MINION_OPTS.copy()
    opts["__role"] = "minion"
    opts["root_dir"] = str(root_dir)
    opts["master_uri"] = "tcp://{ip}:{port}".format(
        ip="127.0.0.1", port=opts["master_port"]
    )
    for name in ("cachedir", "pki_dir", "sock_dir", "conf_dir"):
        dirpath = root_dir / name
        dirpath.mkdir(parents=True)
        opts[name] = str(dirpath)
    opts["log_file"] = "logs/minion.log"
    opts["conf_file"] = os.path.join(opts["conf_dir"], "minion")
    opts["fips_mode"] = FIPS_TESTRUN
    opts["encryption_algorithm"] = "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1"
    opts["signing_algorithm"] = "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
    return opts


@pytest.fixture
def master_opts(tmp_path):
    """
    Default master configuration with relative temporary paths to not require root permissions.
    """
    root_dir = tmp_path / "master"
    opts = salt.config.DEFAULT_MASTER_OPTS.copy()
    opts["__role"] = "master"
    opts["root_dir"] = str(root_dir)
    for name in ("cachedir", "pki_dir", "sock_dir", "conf_dir"):
        dirpath = root_dir / name
        dirpath.mkdir(parents=True)
        opts[name] = str(dirpath)
    opts["log_file"] = "logs/master.log"
    opts["conf_file"] = os.path.join(opts["conf_dir"], "master")
    opts["fips_mode"] = FIPS_TESTRUN
    opts["publish_signing_algorithm"] = (
        "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
    )
    return opts


@pytest.fixture
def syndic_opts(tmp_path):
    """
    Default master configuration with relative temporary paths to not require root permissions.
    """
    root_dir = tmp_path / "syndic"
    opts = salt.config.DEFAULT_MINION_OPTS.copy()
    opts["syndic_master"] = "127.0.0.1"
    opts["__role"] = "minion"
    opts["root_dir"] = str(root_dir)
    for name in ("cachedir", "pki_dir", "sock_dir", "conf_dir"):
        dirpath = root_dir / name
        dirpath.mkdir(parents=True)
        opts[name] = str(dirpath)
    opts["log_file"] = "logs/syndic.log"
    opts["conf_file"] = os.path.join(opts["conf_dir"], "syndic")
    return opts



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.deb_apache as deb_apache
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {deb_apache: {}}


# 'check_site_enabled' function tests: 3


def test_check_site_enabled():
    """
    Test if the specific Site symlink is enabled.
    """
    with patch("os.path.islink", MagicMock(return_value=True)):
        assert deb_apache.check_site_enabled("saltstack.com")


def test_check_site_enabled_default():
    """
    Test if the specific Site symlink is enabled.
    """
    with patch("os.path.islink", MagicMock(side_effect=[False, True])):
        assert deb_apache.check_site_enabled("default")


def test_check_site_enabled_false():
    """
    Test if the specific Site symlink is enabled.
    """
    with patch("os.path.islink", MagicMock(return_value=False)):
        assert not deb_apache.check_site_enabled("saltstack.com")


# 'a2ensite' function tests: 4


def test_a2ensite_notfound():
    """
    Test if it runs a2ensite for the given site.
    """
    mock = MagicMock(return_value=1)
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert deb_apache.a2ensite("saltstack.com") == {
            "Name": "Apache2 Enable Site",
            "Site": "saltstack.com",
            "Status": "Site saltstack.com Not found",
        }


def test_a2ensite_enabled():
    """
    Test if it runs a2ensite for the given site.
    """
    mock = MagicMock(return_value=0)
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert deb_apache.a2ensite("saltstack.com") == {
            "Name": "Apache2 Enable Site",
            "Site": "saltstack.com",
            "Status": "Site saltstack.com enabled",
        }


def test_a2ensite():
    """
    Test if it runs a2ensite for the given site.
    """
    mock = MagicMock(return_value=2)
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert deb_apache.a2ensite("saltstack.com") == {
            "Name": "Apache2 Enable Site",
            "Site": "saltstack.com",
            "Status": 2,
        }


def test_a2ensite_exception():
    """
    Test if it runs a2ensite for the given site.
    """
    mock = MagicMock(side_effect=Exception("error"))
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert str(deb_apache.a2ensite("saltstack.com")) == "error"


# 'a2dissite' function tests: 4


def test_a2dissite_notfound():
    """
    Test if it runs a2dissite for the given site.
    """
    mock = MagicMock(return_value=256)
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert deb_apache.a2dissite("saltstack.com") == {
            "Name": "Apache2 Disable Site",
            "Site": "saltstack.com",
            "Status": "Site saltstack.com Not found",
        }


def test_a2dissite_disabled():
    """
    Test if it runs a2dissite for the given site.
    """
    mock = MagicMock(return_value=0)
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert deb_apache.a2dissite("saltstack.com") == {
            "Name": "Apache2 Disable Site",
            "Site": "saltstack.com",
            "Status": "Site saltstack.com disabled",
        }


def test_a2dissite():
    """
    Test if it runs a2dissite for the given site.
    """
    mock = MagicMock(return_value=2)
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert deb_apache.a2dissite("saltstack.com") == {
            "Name": "Apache2 Disable Site",
            "Site": "saltstack.com",
            "Status": 2,
        }


def test_a2dissite_exception():
    """
    Test if it runs a2dissite for the given site.
    """
    mock = MagicMock(side_effect=Exception("error"))
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert str(deb_apache.a2dissite("saltstack.com")) == "error"


# 'check_mod_enabled' function tests: 2


def test_check_mod_enabled():
    """
    Test if the specific mod symlink is enabled.
    """
    with patch("os.path.islink", MagicMock(return_value=True)):
        assert deb_apache.check_mod_enabled("status.conf")


def test_check_mod_enabled_false():
    """
    Test if the specific mod symlink is enabled.
    """
    with patch("os.path.islink", MagicMock(return_value=False)):
        assert not deb_apache.check_mod_enabled("status.conf")


# 'a2enmod' function tests: 4


def test_a2enmod_notfound():
    """
    Test if it runs a2enmod for the given module.
    """
    mock = MagicMock(return_value=1)
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert deb_apache.a2enmod("vhost_alias") == {
            "Name": "Apache2 Enable Mod",
            "Mod": "vhost_alias",
            "Status": "Mod vhost_alias Not found",
        }


def test_a2enmod_enabled():
    """
    Test if it runs a2enmod for the given module.
    """
    mock = MagicMock(return_value=0)
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert deb_apache.a2enmod("vhost_alias") == {
            "Name": "Apache2 Enable Mod",
            "Mod": "vhost_alias",
            "Status": "Mod vhost_alias enabled",
        }


def test_a2enmod():
    """
    Test if it runs a2enmod for the given module.
    """
    mock = MagicMock(return_value=2)
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert deb_apache.a2enmod("vhost_alias") == {
            "Name": "Apache2 Enable Mod",
            "Mod": "vhost_alias",
            "Status": 2,
        }


def test_a2enmod_exception():
    """
    Test if it runs a2enmod for the given module.
    """
    mock = MagicMock(side_effect=Exception("error"))
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert str(deb_apache.a2enmod("vhost_alias")) == "error"


# 'a2dismod' function tests: 4


def test_a2dismod_notfound():
    """
    Test if it runs a2dismod for the given module.
    """
    mock = MagicMock(return_value=256)
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert deb_apache.a2dismod("vhost_alias") == {
            "Name": "Apache2 Disable Mod",
            "Mod": "vhost_alias",
            "Status": "Mod vhost_alias Not found",
        }


def test_a2dismod_disabled():
    """
    Test if it runs a2dismod for the given module.
    """
    mock = MagicMock(return_value=0)
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert deb_apache.a2dismod("vhost_alias") == {
            "Name": "Apache2 Disable Mod",
            "Mod": "vhost_alias",
            "Status": "Mod vhost_alias disabled",
        }


def test_a2dismod():
    """
    Test if it runs a2dismod for the given module.
    """
    mock = MagicMock(return_value=2)
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert deb_apache.a2dismod("vhost_alias") == {
            "Name": "Apache2 Disable Mod",
            "Mod": "vhost_alias",
            "Status": 2,
        }


def test_a2dismod_exception():
    """
    Test if it runs a2dismod for the given module.
    """
    mock = MagicMock(side_effect=Exception("error"))
    with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
        assert str(deb_apache.a2dismod("vhost_alias")) == "error"


# 'check_conf_enabled' function tests: 2


def test_check_conf_enabled():
    """
    Test if the specific conf symlink is enabled.
    """
    with patch("os.path.islink", MagicMock(return_value=True)):
        assert deb_apache.check_conf_enabled("security.conf")


def test_check_conf_enabled_false():
    """
    Test if the specific conf symlink is enabled.
    """
    with patch("os.path.islink", MagicMock(return_value=False)):
        assert not deb_apache.check_conf_enabled("security.conf")


# 'a2enconf' function tests: 4


def test_a2enconf_notfound():
    """
    Test if it runs a2enconf for the given conf.
    """
    with patch("salt.utils.path.which", MagicMock(return_value="a2enconf")):
        mock = MagicMock(return_value=1)
        with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
            assert deb_apache.a2enconf("security") == {
                "Name": "Apache2 Enable Conf",
                "Conf": "security",
                "Status": "Conf security Not found",
            }


def test_a2enconf_enabled():
    """
    Test if it runs a2enconf for the given conf.
    """
    with patch("salt.utils.path.which", MagicMock(return_value="a2enconf")):
        mock = MagicMock(return_value=0)
        with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
            assert deb_apache.a2enconf("security") == {
                "Name": "Apache2 Enable Conf",
                "Conf": "security",
                "Status": "Conf security enabled",
            }


def test_a2enconf():
    """
    Test if it runs a2enconf for the given conf.
    """
    with patch("salt.utils.path.which", MagicMock(return_value="a2enconf")):
        mock = MagicMock(return_value=2)
        with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
            assert deb_apache.a2enconf("security") == {
                "Name": "Apache2 Enable Conf",
                "Conf": "security",
                "Status": 2,
            }


def test_a2enconf_exception():
    """
    Test if it runs a2enconf for the given conf.
    """
    with patch("salt.utils.path.which", MagicMock(return_value="a2enconf")):
        mock = MagicMock(side_effect=Exception("error"))
        with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
            assert str(deb_apache.a2enconf("security")) == "error"


# 'a2disconf' function tests: 4


def test_a2disconf_notfound():
    """
    Test if it runs a2disconf for the given conf.
    """
    with patch("salt.utils.path.which", MagicMock(return_value="a2disconf")):
        mock = MagicMock(return_value=256)
        with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
            assert deb_apache.a2disconf("security") == {
                "Name": "Apache2 Disable Conf",
                "Conf": "security",
                "Status": "Conf security Not found",
            }


def test_a2disconf_disabled():
    """
    Test if it runs a2disconf for the given conf.
    """
    with patch("salt.utils.path.which", MagicMock(return_value="a2disconf")):
        mock = MagicMock(return_value=0)
        with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
            assert deb_apache.a2disconf("security") == {
                "Name": "Apache2 Disable Conf",
                "Conf": "security",
                "Status": "Conf security disabled",
            }


def test_a2disconf():
    """
    Test if it runs a2disconf for the given conf.
    """
    with patch("salt.utils.path.which", MagicMock(return_value="a2disconf")):
        mock = MagicMock(return_value=2)
        with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
            assert deb_apache.a2disconf("security") == {
                "Name": "Apache2 Disable Conf",
                "Conf": "security",
                "Status": 2,
            }


def test_a2disconf_exception():
    """
    Test if it runs a2disconf for the given conf.
    """
    with patch("salt.utils.path.which", MagicMock(return_value="a2disconf")):
        mock = MagicMock(side_effect=Exception("error"))
        with patch.dict(deb_apache.__salt__, {"cmd.retcode": mock}):
            assert str(deb_apache.a2disconf("security")) == "error"



import urllib.parse

import pytest

import salt.utils.json
import salt.utils.yaml
from tests.support.mock import patch

pytestmark = [
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
]


@pytest.fixture
def app(app):
    app.wsgi_application.config["global"]["tools.hypermedia_in.on"] = True
    return app


@pytest.fixture
def token(http_client, auth_creds, content_type_map, io_loop):
    response = io_loop.run_sync(
        lambda: http_client.fetch(
            "/login",
            method="POST",
            body=urllib.parse.urlencode(auth_creds),
            headers={"Content-Type": content_type_map["form"]},
        )
    )
    assert response.code == 200
    return response.headers["X-Auth-Token"]


@pytest.fixture
def client_headers(token, content_type_map):
    return {
        "Accept": content_type_map["json"],
        "X-Auth-Token": token,
        "Content-Type": content_type_map["form"],
    }


async def test_urlencoded_ctype(http_client, client_headers, content_type_map):
    low = {"client": "local", "fun": "test.ping", "tgt": "jerry"}
    body = urllib.parse.urlencode(low)
    client_headers["Content-Type"] = content_type_map["form"]

    with patch(
        "salt.client.LocalClient.run_job",
        return_value={"jid": "20131219215650131543", "minions": ["jerry"]},
    ):
        # We don't really want to run the job, hence the patch
        response = await http_client.fetch(
            "/", method="POST", body=body, headers=client_headers
        )
    assert response.code == 200
    assert response.body == '{"return": [{"jerry": false}]}'


async def test_json_ctype(http_client, client_headers, content_type_map):
    low = {"client": "local", "fun": "test.ping", "tgt": "jerry"}
    body = salt.utils.json.dumps(low)
    client_headers["Content-Type"] = content_type_map["json"]

    with patch(
        "salt.client.LocalClient.run_job",
        return_value={"jid": "20131219215650131543", "minions": ["jerry"]},
    ):
        # We don't really want to run the job, hence the patch
        response = await http_client.fetch(
            "/", method="POST", body=body, headers=client_headers
        )
    assert response.code == 200
    assert response.body == '{"return": [{"jerry": false}]}'


async def test_json_as_text_out(http_client, client_headers):
    """
    Some service send JSON as text/plain for compatibility purposes
    """
    low = {"client": "local", "fun": "test.ping", "tgt": "jerry"}
    body = salt.utils.json.dumps(low)
    client_headers["Content-Type"] = "text/plain"

    with patch(
        "salt.client.LocalClient.run_job",
        return_value={"jid": "20131219215650131543", "minions": ["jerry"]},
    ):
        # We don't really want to run the job, hence the patch
        response = await http_client.fetch(
            "/", method="POST", body=body, headers=client_headers
        )
    assert response.code == 200
    assert response.body == '{"return": [{"jerry": false}]}'


async def test_yaml_ctype(http_client, client_headers, content_type_map):
    low = {"client": "local", "fun": "test.ping", "tgt": "jerry"}
    body = salt.utils.yaml.safe_dump(low)
    client_headers["Content-Type"] = content_type_map["yaml"]

    with patch(
        "salt.client.LocalClient.run_job",
        return_value={"jid": "20131219215650131543", "minions": ["jerry"]},
    ):
        # We don't really want to run the job, hence the patch
        response = await http_client.fetch(
            "/", method="POST", body=body, headers=client_headers
        )
    assert response.code == 200
    assert response.body == '{"return": [{"jerry": false}]}'
